import unittest

import collections

from alphatwirl.loop import Collector

##__________________________________________________________________||
class MockResultsCombinationMethod(object):
    def __init__(self, ret = None):
        self.ret = ret

    def combine(self, dataset_readers_list):
        self.args = dataset_readers_list
        return self.ret

##__________________________________________________________________||
class MockDeliveryMethod(object):
    def __init__(self):
        self.results = None

    def deliver(self, results):
        self.results = results

##__________________________________________________________________||
MockResult = collections.namedtuple('MockResult', 'name')

##__________________________________________________________________||
MockArgs = collections.namedtuple('MockArgs', 'name')

##__________________________________________________________________||
class TestCollector(unittest.TestCase):

    def setUp(self):

        self.result = MockResult('result1')
        self.resultsCombinationMethod = MockResultsCombinationMethod(self.result)
        self.deliveryMethod = MockDeliveryMethod()
        self.obj = Collector(self.resultsCombinationMethod, self.deliveryMethod)

    def test_repr(self):
        repr(self.obj)

    def test_collect(self):

        args = MockArgs('args1')
        self.assertIs(self.result, self.obj.collect(args))
        self.assertIs(args, self.resultsCombinationMethod.args)
        self.assertIs(self.result, self.deliveryMethod.results)

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import os
import sys
import pytest

has_no_ROOT = False
try:
    import ROOT
except ImportError:
    has_no_ROOT = True

if not has_no_ROOT:
    from alphatwirl.roottree import BEvents as Events
    from alphatwirl.roottree import Branch

##__________________________________________________________________||
pytestmark = pytest.mark.skipif(has_no_ROOT, reason="has no ROOT")

##__________________________________________________________________||
@pytest.fixture()
def tree():
    input_file_name = 'sample_04.root'
    tree_name = 'tree'
    input_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', input_file_name)

    input_file = ROOT.TFile.Open(input_path)
    yield input_file.Get(tree_name)

@pytest.fixture()
def events(tree):
    yield Events(tree)

def test_event(events):

    event = events[0]
    assert [-125, 38] == list(event.bChar)
    assert [253, 20] == list(event.bUChar)
    assert [-10, 120] == list(event.bShort)
    assert [65530, 21221] == list(event.bUShort)
    assert [-2147483626, 512] == list(event.bInt)
    assert [4294967290, 1253] == list(event.bUInt)
    assert [-0.123, 42.344] == pytest.approx(list(event.bFloat), rel=1e-6)
    assert [-2.345, 51.224] == list(event.bDouble)
    assert [-4611686018427387900, 12345] == list(event.bLong64)
    assert [9223372036854775802, 12345514] == list(event.bULong64)
    assert [1, 1] == list(event.bBool)

##__________________________________________________________________||



from .BProgressMonitor import BProgressMonitor
from .NullProgressMonitor import NullProgressMonitor
from .ProgressBar import ProgressBar
from .ProgressMonitor import ProgressMonitor, Queue
from .ProgressPrint import ProgressPrint
from .ProgressReportPickup import ProgressReportPickup
from .ProgressReport import ProgressReport
from .ProgressReporter import ProgressReporter
from .main import atpbar

##__________________________________________________________________||
try:
    from .ProgressBarJupyter import ProgressBarJupyter
except ImportError:
    pass

##__________________________________________________________________||
_progress_reporter = None

def report_progress(report):
    if _progress_reporter is None:
        return
    _progress_reporter.report(report)

##__________________________________________________________________||



import copy

##__________________________________________________________________||
class MockReader(object):
    def __init__(self, summarizer):
        self.summarizer = summarizer

    def results(self):
        return self.summarizer

    def __repr__(self):
        return '{}(summarizer = {!r})'.format(
            self.__class__.__name__,
            self.summarizer
        )

##__________________________________________________________________||
class MockSummarizer(object):
    def __init__(self, results):
        self._results = results

    def __repr__(self):
        return '{}({!r})'.format(
            self.__class__.__name__,
            self._results
        )

    def __add__(self, other):
        if other == 0:
            res = copy.copy(self._results)
        else:
            res = copy.copy(self._results) + copy.copy(other._results)
        return self.__class__(res)

    def __radd__(self, other):
        return self.__add__(other)

    def to_tuple_list(self):
        return self._results

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>

import numpy as np
import copy

##__________________________________________________________________||
class Count(object):
    """
    Args:
        val : If None, initialize with 0. i.e., not counted. Otherwise,
              typically an empty tuple, counted with the weight. Ignored
              if contents are given.
        weight (float) : The weight
        contents : Specified contents unless None
    """

    def __init__(self, val=None, weight=1, contents=None):

        if contents is not None:
            ## self.contents = copy.deepcopy(contents)
            self.contents = contents
            return

        if val is None:
            self.contents = [np.array((0, 0))]
            return

        self.contents = [np.array((weight, weight**2))]

    def __add__(self, other):
        contents = [self.contents[0] + other.contents[0]]
        return self.__class__(contents=contents)

    def __radd__(self, other):
        # is called with other=0 when e.g. sum([obj1, obj2])
        if other == 0:
            return self.__class__() + self
        raise TypeError('unsupported: {!r} + {!r}'.format(other, self))

    def __repr__(self):
        return '{}(contents={!r})'.format(self.__class__.__name__, self.contents)

    def __eq__(self, other):
        if len(self.contents) != len(other.contents):
            return False
        cmps = [np.all(self.contents[i] == other.contents[i]) for i in range(len(self.contents))]
        return all(cmps)

    def __copy__(self):
        contents = [np.copy(self.contents[0])]
        return self.__class__(contents=contents)

##__________________________________________________________________||




#!/usr/bin/env python
# Tai Sakuma <tai.sakuma@gmail.com>
from __future__ import print_function
import os, sys
import argparse
import tarfile
import gzip

try:
   import cPickle as pickle
except:
   import pickle

##__________________________________________________________________||
parser = argparse.ArgumentParser()
parser.add_argument('pickle', help = 'path to a pickle file')
args = parser.parse_args()


##__________________________________________________________________||
pickle_path = os.path.abspath(args.pickle)

##__________________________________________________________________||
os.chdir(os.path.dirname(os.path.dirname(os.path.dirname(pickle_path))))

##__________________________________________________________________||
dirname = 'python_modules'
tarname = dirname + '.tar.gz'
if os.path.exists(tarname) and not os.path.exists(dirname):
    tar = tarfile.open(tarname)
    tar.extractall()
    tar.close()
sys.path.insert(0, dirname)

##__________________________________________________________________||
f = gzip.open(pickle_path, 'rb')
package = pickle.load(f)

print(package)

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>

import pandas as pd

from .ToTupleList import ToTupleList

##__________________________________________________________________||
class ToDataFrame(object):
    def __init__(self, summaryColumnNames):

        self.summaryColumnNames = summaryColumnNames
        self.to_tuple_list = ToTupleList(summaryColumnNames = summaryColumnNames)

    def __repr__(self):

        name_value_pairs = (
            ('summaryColumnNames', self.summaryColumnNames),
        )
        return '{}({})'.format(
            self.__class__.__name__,
            ', '.join(['{} = {!r}'.format(n, v) for n, v in name_value_pairs]),
        )

    def combine(self, dataset_readers_list):
        tuple_list = self.to_tuple_list.combine(dataset_readers_list)
        if tuple_list is None:
            return None
        header = tuple_list[0]
        contents = tuple_list[1:]
        return pd.DataFrame(contents, columns = header)

##__________________________________________________________________||



import alphatwirl.summary as summary
import unittest

##__________________________________________________________________||
class MockBinningPlusOneNext(object):
    def __call__(self, val):
        return val

    def next(self, val):
        return val + 1

##__________________________________________________________________||
class MockBinningNoneNext(object):
    def __call__(self, val):
        return val

    def next(self, val):
        return None

##__________________________________________________________________||
class MockBinningSameNext(object):
    def __call__(self, val):
        return val

    def next(self, val):
        return val

##__________________________________________________________________||
class TestNextKeyComposer(unittest.TestCase):

    def test_repr(self):
        binnings = (MockBinningPlusOneNext(), MockBinningPlusOneNext(), MockBinningPlusOneNext())
        keyComposer = summary.NextKeyComposer(binnings)
        repr(keyComposer)

    def test_call(self):
        binnings = (MockBinningPlusOneNext(), MockBinningPlusOneNext(), MockBinningPlusOneNext())
        keyComposer = summary.NextKeyComposer(binnings)
        self.assertEqual(((12, 8, 20), (11, 9, 20), (11, 8, 21)), keyComposer((11, 8, 20)))

    def test_call_one_none(self):
        binnings = (MockBinningPlusOneNext(), MockBinningNoneNext(), MockBinningPlusOneNext())
        keyComposer = summary.NextKeyComposer(binnings)
        self.assertEqual(((12, 8, 20), (11, 8, 21)), keyComposer((11, 8, 20)))

    def test_call_all_none(self):
        binnings = (MockBinningNoneNext(), MockBinningNoneNext(), MockBinningNoneNext())
        keyComposer = summary.NextKeyComposer(binnings)
        self.assertEqual(( ), keyComposer((11, 8, 20)))

    def test_call_no_binning(self):
        binnings = ( )
        keyComposer = summary.NextKeyComposer(binnings)
        self.assertEqual(( ), keyComposer(( )))

    def test_call_one_same(self):
        binnings = (MockBinningPlusOneNext(), MockBinningSameNext(), MockBinningPlusOneNext())
        keyComposer = summary.NextKeyComposer(binnings)
        self.assertEqual(((12, 8, 20), (11, 8, 21)), keyComposer((11, 8, 20)))

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import sys
import logging

import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

has_no_ROOT = False
try:
    import ROOT
except ImportError:
    has_no_ROOT = True

if not has_no_ROOT:
    from alphatwirl.roottree import Branch
    from alphatwirl.roottree import BranchBuilder
    from alphatwirl.roottree import BranchAddressManager
    from alphatwirl.roottree import BranchAddressManagerForVector

##__________________________________________________________________||
pytestmark = pytest.mark.skipif(has_no_ROOT, reason="has no ROOT")

##__________________________________________________________________||
class MockLeaf(object):
    def __init__(self, name, typename):
        self.name = name
        self.typename = typename

    def GetName(self):
        return self.name

    def GetTypeName(self):
        return self.typename

class MockTree(object):
    def __init__(self, *_, **__):
        self.leaf_names = ('run', 'evt', 'njet', 'jet_pt', 'met_pt', 'trigger_path', 'EventAuxiliary')
        self.leaf_type_names = ('Int_t', 'Int_t', 'Int_t', 'Double_t', 'Double_t', 'vector<string>', 'edm::EventAuxiliary')
        self.leafs = dict([(n, MockLeaf(n, t)) for n, t in zip(self.leaf_names, self.leaf_type_names)])

    def SetBranchStatus(self, *_, **__):
        pass

    def GetListOfLeaves(self):
        return self.leafs.values()

    def GetLeaf(self, name):
        return self.leafs[name]

@pytest.fixture()
def mockTree():
    ret = mock.Mock(wraps=MockTree())
    return ret

##__________________________________________________________________||
@pytest.fixture(autouse=True)
def mockBranchAddressManager(monkeypatch):
    ret = mock.Mock(spec=BranchAddressManager)
    def getArrays(tree, branchName):
        leafNames = ('run', 'evt', 'njet', 'jet_pt', 'met_pt')
        if branchName in leafNames:
            return mock.Mock(name=branchName), mock.Mock(name='{}_counter'.format(branchName))
        return None, None
    ret.getArrays.side_effect = getArrays
    module = sys.modules['alphatwirl.roottree.BranchBuilder']
    monkeypatch.setattr(module, 'branchAddressManager', ret)
    yield ret

@pytest.fixture(autouse=True)
def mockBranchAddressManagerForVector(monkeypatch):
    ret = mock.Mock(spec=BranchAddressManagerForVector)
    def getVector(tree, branchName):
        leaf_dict = dict(trigger_path='string')
        if branchName in leaf_dict:
            return mock.Mock(name=branchName, spec=ROOT.vector(leaf_dict[branchName]))
        return None
    ret.getVector.side_effect = getVector
    module = sys.modules['alphatwirl.roottree.BranchBuilder']
    monkeypatch.setattr(module, 'branchAddressManagerForVector', ret)
    yield ret

##__________________________________________________________________||
def test_repr():
    obj = BranchBuilder()
    repr(obj)

def test_register_tree(mockTree):
    obj = BranchBuilder()
    obj.register_tree(mockTree)
    assert [mock.call('*', 0)] == mockTree.SetBranchStatus.call_args_list
    assert mockTree in obj.itsdict

def test_register_tree_two_builders(mockTree):
    obj1 = BranchBuilder()
    obj2 = BranchBuilder()
    obj1.register_tree(mockTree)
    obj2.register_tree(mockTree)
    assert 1 == mockTree.SetBranchStatus.call_count

def test_call_ctypes(mockTree):
    obj = BranchBuilder()
    obj.register_tree(mockTree)
    result = obj(mockTree, 'jet_pt')
    assert isinstance(result, Branch)

def test_call_not_registered(mockTree, caplog):
    obj = BranchBuilder()
    with caplog.at_level(logging.WARNING):
        result = obj(mockTree, 'jet_pt')
    assert isinstance(result, Branch)
    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'WARNING'
    assert 'BranchBuilder' in caplog.records[0].name
    assert 'tree is not registered' in caplog.records[0].msg

def test_call_ctypes_same_objects_different_calls(mockTree):
    obj = BranchBuilder()
    obj.register_tree(mockTree)
    result1 = obj(mockTree, 'jet_pt')
    result2 = obj(mockTree, 'jet_pt')
    assert result1 is result2

def test_call_ctypes_same_objects_different_builders(mockTree):
    obj1 = BranchBuilder()
    obj2 = BranchBuilder()
    obj1.register_tree(mockTree)
    obj2.register_tree(mockTree)
    result1 = obj1(mockTree, 'jet_pt')
    result2 = obj2(mockTree, 'jet_pt')
    assert result1 is result2

@pytest.mark.skipif(sys.version_info[0]!=2, reason="skip for Python 3")
def test_call_stdvector(mockTree):
    obj = BranchBuilder()
    obj.register_tree(mockTree)
    result = obj(mockTree, 'trigger_path')
    assert isinstance(result, ROOT.vector('string'))

def test_call_unknown_type(mockTree, caplog):
    obj = BranchBuilder()
    obj.register_tree(mockTree)
    with caplog.at_level(logging.WARNING):
        result = obj(mockTree, 'EventAuxiliary')
    assert result is None
    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'WARNING'
    assert 'BranchBuilder' in caplog.records[0].name
    assert 'unknown leaf type' in caplog.records[0].msg

def test_call_no_such_branch(mockTree):
    obj = BranchBuilder()
    obj.register_tree(mockTree)
    result = obj(mockTree, 'no_such_branch')
    assert result is None

##__________________________________________________________________||
def test_empty_chain(caplog):
    tree_name = 'tree'
    chain = ROOT.TChain(tree_name)
    # add no files to the chain

    obj = BranchBuilder()
    obj.register_tree(chain)

    with caplog.at_level(logging.WARNING):
        assert obj(chain, 'var') is None

    assert len(caplog.records) == 2
    assert caplog.records[0].levelname == 'WARNING'
    assert 'BranchBuilder' in caplog.records[0].name
    assert 'cannot get' in caplog.records[0].msg

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import logging
import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

from alphatwirl.loop import EventLoopRunner

##__________________________________________________________________||
@pytest.fixture()
def obj():
    return EventLoopRunner()

##__________________________________________________________________||
def test_repr(obj):
    repr(obj)

def test_begin(obj):
    obj.begin()

def test_end(obj):
    obj.begin()
    assert [ ] == obj.end()

def test_run(obj):
    obj.begin()

    result1 = mock.Mock(name='result1')
    eventLoop1 = mock.Mock(name='eventLoop1')
    eventLoop1.return_value = result1
    assert 0 == obj.run(eventLoop1)

    assert [mock.call()] == eventLoop1.call_args_list

    result2 = mock.Mock(name='result2')
    eventLoop2 = mock.Mock(name='eventLoop2')
    eventLoop2.return_value = result2
    assert 1 == obj.run(eventLoop2)

    assert [mock.call()] == eventLoop2.call_args_list

    assert [result1, result2] == obj.end()

def test_run_multiple(obj):
    obj.begin()

    result1 = mock.Mock(name='result1')
    eventLoop1 = mock.Mock(name='eventLoop1')
    eventLoop1.return_value = result1

    result2 = mock.Mock(name='result2')
    eventLoop2 = mock.Mock(name='eventLoop2')
    eventLoop2.return_value = result2

    assert [0, 1] == obj.run_multiple([eventLoop1, eventLoop2])

    assert [mock.call()] == eventLoop1.call_args_list
    assert [mock.call()] == eventLoop2.call_args_list

    assert [result1, result2] == obj.end()
    assert [ ] == obj.end()

##__________________________________________________________________||
def test_receive(obj):
    obj.begin()

    result1 = mock.Mock(name='result1')
    eventLoop1 = mock.Mock(name='eventLoop1')
    eventLoop1.return_value = result1

    result2 = mock.Mock(name='result2')
    eventLoop2 = mock.Mock(name='eventLoop2')
    eventLoop2.return_value = result2

    assert [0, 1] == obj.run_multiple([eventLoop1, eventLoop2])

    assert [mock.call()] == eventLoop1.call_args_list
    assert [mock.call()] == eventLoop2.call_args_list

    assert [(0, result1), (1, result2)] == obj.receive()
    assert [ ] == obj.receive()

def test_poll(obj):
    obj.begin()

    result1 = mock.Mock(name='result1')
    eventLoop1 = mock.Mock(name='eventLoop1')
    eventLoop1.return_value = result1

    result2 = mock.Mock(name='result2')
    eventLoop2 = mock.Mock(name='eventLoop2')
    eventLoop2.return_value = result2

    assert [0, 1] == obj.run_multiple([eventLoop1, eventLoop2])

    assert [mock.call()] == eventLoop1.call_args_list
    assert [mock.call()] == eventLoop2.call_args_list

    assert [(0, result1), (1, result2)] == obj.poll()
    assert [ ] == obj.poll()

def test_receive_one(obj):
    obj.begin()

    result1 = mock.Mock(name='result1')
    eventLoop1 = mock.Mock(name='eventLoop1')
    eventLoop1.return_value = result1

    result2 = mock.Mock(name='result2')
    eventLoop2 = mock.Mock(name='eventLoop2')
    eventLoop2.return_value = result2

    assert [0, 1] == obj.run_multiple([eventLoop1, eventLoop2])

    assert [mock.call()] == eventLoop1.call_args_list
    assert [mock.call()] == eventLoop2.call_args_list

    assert (0, result1) == obj.receive_one()
    assert (1, result2) == obj.receive_one()
    assert obj.receive_one() is None

##__________________________________________________________________||




# Tai Sakuma <tai.sakuma@gmail.com>

##__________________________________________________________________||
class LambdaStr(object):
    """select events to which a lambda returns True.

    A lambda should be given as a string to ``__init__`` and will be
    evaluated in ``begin()``. This is because a lambda is not
    picklable.

    In the multiprocessing mode, ``__init__()`` is called in the main
    process. Then, the instance will be pickled and sent to
    subprocesses. ``begin()`` will be called in the subprocesses.

    """
    def __init__(self, lambda_str, name=None):
        self.name = name if name is not None else lambda_str
        self.lambda_str = lambda_str

    def __repr__(self):
        return '{}(name={!r}, lambda_str={!r})'.format(
            self.__class__.__name__,
            self.name,
            self.lambda_str
        )

    def begin(self, event):
        self.func = eval('lambda ' + self.lambda_str)

    def event(self, event):
        try:
            return self.func(event)
        except BaseException as e:
            raise ValueError('{}: error in executing "{}". raised -- {}: {}'.format(
                self.__class__.__name__,
                self.lambda_str, type(e).__name__, e
            ))

    def __call__(self, event):
        return self.event(event)

    def end(self):
        self.func = None

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
from collections import OrderedDict

import pytest

from alphatwirl.loop.merge import merge_in_order

##__________________________________________________________________||
class MockData(object):
    def __init__(self, name):
        self.name = name
        self.merged = [ ]

    def __repr__(self):
        name_value_pairs = (
            ('name', self.name),
            ('merged', self.merged),
        )
        return '{}({})'.format(
            self.__class__.__name__,
            ', '.join(['{}={!r}'.format(n, v) for n, v in name_value_pairs]),
        )

    def merge(self, data_):
        self.merged.extend([data_.name] + data_.merged)


@pytest.mark.parametrize('orders', [
    [0, 1, 2, 3, 4, 5],
    [5, 4, 3, 2, 1, 0],
    [1, 2, 4, 3, 0, 5],
])
def test_merge_in_order_six_data(orders):
    ndata = 6
    map_ = OrderedDict([(i, None) for i in range(ndata)])

    data = [MockData(name='data{}'.format(i)) for i in range(ndata)]

    for i in orders:
        merge_in_order(map_, i, data[i])

    assert ['data1', 'data2', 'data3', 'data4', 'data5'] == data[0].merged

@pytest.mark.parametrize('orders', [
    [0, 1, 2, 3, 4, 5],
    [5, 4, 3, 2, 1, 0],
    [1, 2, 4, 3, 0, 5],
])
def test_merge_in_order_six_data_id_shifted(orders):
    ndata = 6
    map_ = OrderedDict([(i + 1000, None) for i in range(ndata)])

    data = [MockData(name='data{}'.format(i)) for i in range(ndata)]

    for i in orders:
        id_ = i + 1000
        merge_in_order(map_, id_, data[i])

    assert ['data1', 'data2', 'data3', 'data4', 'data5'] == data[0].merged

def test_merge_in_order_one_data():
    ndata = 1
    map_ = OrderedDict([(1000, None)])
    data = [MockData(name='data0')]
    merge_in_order(map_, 1000, data[0])

    assert [ ] == data[0].merged

##__________________________________________________________________||
class MockDataWoMerge(object):
    def __init__(self, name):
        self.name = name
        self.merged = [ ]

    def __repr__(self):
        name_value_pairs = (
            ('name', self.name),
            ('merged', self.merged),
        )
        return '{}({})'.format(
            self.__class__.__name__,
            ', '.join(['{}={!r}'.format(n, v) for n, v in name_value_pairs]),
        )

@pytest.mark.parametrize('orders', [
    [0, 1, 2, 3, 4, 5],
    [5, 4, 3, 2, 1, 0],
    [1, 2, 4, 3, 0, 5],
])
def test_merge_in_order_no_merge_method(orders):
    ndata = 6
    map_ = OrderedDict([(i, None) for i in range(ndata)])

    data = [MockDataWoMerge(name='data{}'.format(i)) for i in range(ndata)]

    for i in orders:
        merge_in_order(map_, i, data[i])

    assert [ ] == data[0].merged

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import itertools
import numbers

##__________________________________________________________________||
class BackrefMultipleArrayReader(object):
    def __init__(self, arrays, idxs_conf, backref_idxs=None):

        # e.g.,
        # arrays = [[ ], [ ], [ ], [ ], [ ], [ ], [ ], [ ]]
        # idxs_conf = (0, '*', None, '*', None, None, None, None)
        # backref_idxs = [None, None, 1, None, 3, 1, 1, 3]

        #
        self._check_args(arrays, idxs_conf, backref_idxs)

        #
        self.take_fast_path = not self._is_backref_used(backref_idxs)

        if self.take_fast_path:
            self._init_fast_path(arrays, idxs_conf)
            return

        self._init_full_path(arrays, idxs_conf, backref_idxs)

    def _check_args(self, arrays, idxs_conf, backref_idxs):

        if not len(arrays) == len(idxs_conf):
            raise ValueError(
                "these two arguments must have the same length: arrays={}, idxs_conf={}".format(
                    arrays, idxs_conf
                )
            )

        if backref_idxs is not None:
            if not len(idxs_conf) == len(backref_idxs):
                raise ValueError(
                    "backref_idxs must have the same length as idxs_conf: idxs_conf={}, backref_idxs={}".format(
                        idxs_conf, backref_idxs
                    )
                )

    def _is_backref_used(self, backref_idxs):
        if backref_idxs is None:
            return False
        return any([e is not None for e in backref_idxs])

    def _init_common(self, arrays, idxs_conf):
        self.arrays = arrays

        self.wildcard_conf = tuple(c == '*' for c in idxs_conf)
        # e.g., (False, True, False, True, False, False, False, False)

        self.idxs_conf = tuple(c if isinstance(c, numbers.Number) else None for c in idxs_conf)
        # e.g., (0, None, None, None, None, None, None, None)

    def _init_fast_path(self, arrays, idxs_conf):
        self._init_common(arrays, idxs_conf)
        self._zipped = list(zip(self.arrays, self.idxs_conf, self.wildcard_conf))

    def _init_full_path(self, arrays, idxs_conf, backref_idxs):
        self._init_common(arrays, idxs_conf)

        self.uniq_idxs = [ ]
        self.ref_idxs = [ ]
        for i in backref_idxs:
            if i is None:
                l = [ ]
                self.uniq_idxs.append(l)
                self.ref_idxs.append(l)
            else:
                self.ref_idxs.append(self.ref_idxs[i])

        # e.g.,
        # self.uniq_idxs = [[], [], []]
        # [id(o) for o in self.uniq_idxs] = [048, 480, 120]
        # self.ref_idxs = [[], [], [], [], [], [], [], []]
        # [id(o) for o in self.ref_idxs] = [048, 480, 480, 120, 120, 480, 480, 120]

        self._zipped_1 = list(zip(self.arrays, self.idxs_conf, self.wildcard_conf, self.ref_idxs))
        self._zipped_2 = list(zip(self.arrays, self.ref_idxs))

    def read(self):
        if self.take_fast_path:
            return self._fast_read_without_backref()
        return self._read_with_backref()

    def _fast_read_without_backref(self):

        # e.g.,
        # self.idxs_conf = (1, None, None, 2, None)
        # self.wildcard_conf = (False, True, True, False, True)
        # self.arrays = [
        #     [55, 66, 77],
        #     [12, 13, 14],
        #     [104, 105],
        #     [222, 333, 444, 555],
        #     [1001, 1002, 1003]
        # ]

        # read arrays
        vals = [ ]
        for array, idx, wild in self._zipped:
            if wild:
                vals.append(list(array))
                continue
            if idx < len(array):
                vals.append([array[idx]])
                continue
            # idx is out of the range
            vals.append([ ])
        # e.g.,
        # vals = [
        #     [66],
        #     [12, 13, 14],
        #     [104, 105],
        #     [444],
        #     [1001, 1002, 1003]
        # ]

        # expand with all combinations
        ret =  tuple(itertools.product(*vals))
        # e.g.,
        # ret = (
        #     (66, 12, 104, 444, 1001),
        #     (66, 12, 104, 444, 1002),
        #     (66, 12, 104, 444, 1003),
        #     (66, 12, 105, 444, 1001),
        #     (66, 12, 105, 444, 1002),
        #     (66, 12, 105, 444, 1003),
        #     (66, 13, 104, 444, 1001),
        #     (66, 13, 104, 444, 1002),
        #     (66, 13, 104, 444, 1003),
        #     (66, 13, 105, 444, 1001),
        #     (66, 13, 105, 444, 1002),
        #     (66, 13, 105, 444, 1003),
        #     (66, 14, 104, 444, 1001),
        #     (66, 14, 104, 444, 1002),
        #     (66, 14, 104, 444, 1003),
        #     (66, 14, 105, 444, 1001),
        #     (66, 14, 105, 444, 1002),
        #     (66, 14, 105, 444, 1003)
        # )

        return ret

    def _read_with_backref(self):

        # e.g.,
        # self.arrays = [
        #     [1001],
        #     [12, 13, 14, 15],
        #     [104, 105, 106, 107],
        #     [51, 52],
        #     [84, 85, 86],
        #     [403, 404, 405],
        #     [207, 208, 209, 210],
        #     [91, 92, 93]
        # ]

        # initialize self.uniq_idxs
        for i in self.uniq_idxs:
            i[:] = [ ]
        # e.g.,
        # self.uniq_idxs = [[], [], []]
        # self.ref_idxs = [[], [], [], [], [], [], [], []]


        # fill self.ref_idxs
        for array, idx, wild, ref in self._zipped_1:
            larray = len(array)
            if idx is not None:
                ref[:] = [idx] if idx < larray else [ ]
                continue
            if wild:
                ref[:] = range(larray)
                continue
            # backref
            ref[:] = [j for j in ref if j < larray]
        # e.g.,
        # self.uniq_idxs = [
        #     [0],
        #     [0, 1, 2],
        #     [0, 1]
        # ]
        # self.ref_idxs = [
        #     [0],
        #     [0, 1, 2],
        #     [0, 1, 2],
        #     [0, 1],
        #     [0, 1],
        #     [0, 1, 2],
        #     [0, 1, 2],
        #     [0, 1]
        # ]

        # expand self.uniq_idxs with all combinations
        prod = tuple(itertools.product(*self.uniq_idxs))
        for i in range(len(self.uniq_idxs)):
            self.uniq_idxs[i][:] = [p[i] for p in prod]
        # e.g.,
        # self.uniq_idxs = [
        #     [0, 0, 0, 0, 0, 0],
        #     [0, 0, 1, 1, 2, 2],
        #     [0, 1, 0, 1, 0, 1]
        # ]
        # self.ref_idxs = [
        #     [0, 0, 0, 0, 0, 0],
        #     [0, 0, 1, 1, 2, 2],
        #     [0, 0, 1, 1, 2, 2],
        #     [0, 1, 0, 1, 0, 1],
        #     [0, 1, 0, 1, 0, 1],
        #     [0, 0, 1, 1, 2, 2],
        #     [0, 0, 1, 1, 2, 2],
        #     [0, 1, 0, 1, 0, 1]
        # ]

        ret = [[a[j] for j in i] for a, i in self._zipped_2]
        # e.g.,
        # ret = [
        #     [1001, 1001, 1001, 1001, 1001, 1001],
        #     [12, 12, 13, 13, 14, 14],
        #     [104, 104, 105, 105, 106, 106],
        #     [51, 52, 51, 52, 51, 52],
        #     [84, 85, 84, 85, 84, 85],
        #     [403, 403, 404, 404, 405, 405],
        #     [207, 207, 208, 208, 209, 209],
        #     [91, 92, 91, 92, 91, 92]
        # ]

        # take transpose
        ret = tuple(map(tuple, zip(*ret)))
        # e.g.,
        # ret = (
        #     (1001, 12, 104, 51, 84, 403, 207, 91),
        #     (1001, 12, 104, 52, 85, 403, 207, 92),
        #     (1001, 13, 105, 51, 84, 404, 208, 91),
        #     (1001, 13, 105, 52, 85, 404, 208, 92),
        #     (1001, 14, 106, 51, 84, 405, 209, 91),
        #     (1001, 14, 106, 52, 85, 405, 209, 92)
        # )
        return ret

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import re

##__________________________________________________________________||
def parse_indices_config(indices):
    indices = list(indices)

    # indices e.g., [None, None, '(*)', '(*)', '\\1', '\\2']

    # replace None with 0
    indices = [0 if i is None else i for i in indices]
    # e.g., [0, 0, '(*)', '(*)', '\\1', '\\2']

    # search for elements in parentheses, e.g. '(*)'
    # at the momentum, only the asterisk '*' can be in the parentheses
    idxRefs = [re.search(r'^\((.*)\)$', i) if isinstance(i, str) else None for i in indices]
    # e.g., [None, None, <Match object>, <Match object>, None, None]

    # remove parentheses
    indices = [r.group(1) if r else i for i, r in zip(indices, idxRefs)]
    # e.g., [0, 0, '*', '*', '\\1', '\\2']

    ref = 1
    for i, v in enumerate(idxRefs):
        if v:
            idxRefs[i] = ref
            ref += 1
    # e.g., idxRefs  = [None, None, 1, 2, None, None]

    backrefIdxs = [int(i[1:]) if isinstance(i, str) and i.startswith('\\') else None for i in indices]
    # e.g., [None, None, None, None, 1, 2] # the original refs

    backrefIdxs = [None if i is None else idxRefs.index(i) for i in backrefIdxs]
    # e.g., [None, None, None, None, 2, 3] # indices in the list "indices"

    # e.g.:
    # backrefIdxs = [None, None, None, None, 2, 3]
    # indices = [0, 0, '*', '*', '\\1', '\\2']
    return backrefIdxs, indices

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import sys
import logging

from alphatwirl import concurrently, progressbar
from alphatwirl.misc.deprecation import _deprecated
from alphatwirl.misc.deprecation import _deprecated_func_option

from .parallel import Parallel

##__________________________________________________________________||
@_deprecated_func_option('htcondor_job_desc_extra', msg='use dispatcher_options instead')
def build_parallel(parallel_mode, quiet=True, processes=4, user_modules=[ ],
                   htcondor_job_desc_extra=[ ], dispatcher_options=dict()):

    dispatchers = ('subprocess', 'htcondor')
    parallel_modes = ('multiprocessing', ) + dispatchers
    default_parallel_mode = 'multiprocessing'

    if not parallel_mode in parallel_modes:
        logger = logging.getLogger(__name__)
        logger.warning('unknown parallel_mode "{}", use default "{}"'.format(
            parallel_mode, default_parallel_mode
        ))
        parallel_mode = default_parallel_mode

    if parallel_mode == 'multiprocessing':
        return _build_parallel_multiprocessing(quiet=quiet, processes=processes)

    ## TODO: to be deleted as htcondor_job_desc_extra is obsolete
    dispatcher_options = dispatcher_options.copy()
    if parallel_mode == 'htcondor':
        if htcondor_job_desc_extra:
            dispatcher_options['job_desc_extra'] = htcondor_job_desc_extra

    return _build_parallel_dropbox(
        parallel_mode=parallel_mode,
        user_modules=user_modules,
        dispatcher_options=dispatcher_options
    )

##__________________________________________________________________||
def _build_parallel_dropbox(parallel_mode, user_modules,
                            dispatcher_options):

    workingarea_topdir = '_ccsp_temp'
    python_modules = set(user_modules)
    python_modules.add('alphatwirl')
    workingarea_options = dict(topdir=workingarea_topdir, python_modules=python_modules)

    if parallel_mode == 'htcondor':
        dispatcher_class = concurrently.HTCondorJobSubmitter
    else:
        dispatcher_class = concurrently.SubprocessRunner

    return _build_parallel_dropbox_(
        workingarea_options, dispatcher_class, dispatcher_options
    )

def _build_parallel_dropbox_(workingarea_options,
                             dispatcher_class, dispatcher_options):

    workingarea = concurrently.WorkingArea(**workingarea_options)

    dispatcher = dispatcher_class(**dispatcher_options)

    dropbox = concurrently.TaskPackageDropbox(
        workingArea=workingarea,
        dispatcher=dispatcher
    )
    communicationChannel = concurrently.CommunicationChannel(
        dropbox=dropbox
    )

    progressMonitor = progressbar.NullProgressMonitor()

    return Parallel(progressMonitor, communicationChannel, workingarea)

##__________________________________________________________________||
def _build_parallel_multiprocessing(quiet, processes):

    if quiet:
        progressBar = None
    elif sys.stdout.isatty():
        progressBar = progressbar.ProgressBar()
    elif is_jupyter_notebook():
        progressBar = progressbar.ProgressBarJupyter()
    else:
        progressBar = progressbar.ProgressPrint()

    if processes is None or processes == 0:
        progressMonitor = progressbar.NullProgressMonitor() if quiet else progressbar.ProgressMonitor(presentation = progressBar)
        communicationChannel = concurrently.CommunicationChannel0()
    else:
        progressMonitor = progressbar.NullProgressMonitor() if quiet else progressbar.BProgressMonitor(presentation = progressBar)
        dropbox = concurrently.MultiprocessingDropbox(processes, progressMonitor)
        communicationChannel = concurrently.CommunicationChannel(dropbox = dropbox)

    return Parallel(progressMonitor, communicationChannel)

##__________________________________________________________________||
def is_jupyter_notebook():
    try:
        from IPython import get_ipython
        if 'IPKernelApp' in get_ipython().config:
            return True
    except:
        pass
    return False

##__________________________________________________________________||
@_deprecated(msg='use alphatwirl.parallel.build.build_parallel() instead.')
def build_parallel_multiprocessing(quiet, processes):
    return build_parallel(
        parallel_mode='multiprocessing',
        quiet=quiet, processes=processes
    )

##__________________________________________________________________||



#!/usr/bin/env python
# Tai Sakuma <tai.sakuma@gmail.com>
from __future__ import print_function

import os, sys
import errno
import argparse
import tarfile
import signal
import gzip
import json
import logging
import cProfile, pstats

from io import StringIO, BytesIO

try:
    import cPickle as pickle
except:
    import pickle

##__________________________________________________________________||
parser = argparse.ArgumentParser()
parser.add_argument('paths', nargs=argparse.REMAINDER, help='paths to task packages')
parser.add_argument('--profile', action='store_true', help='run profile')
parser.add_argument('--profile-out-path', default=None, help='path to write the result of profile')
args = parser.parse_args()

##__________________________________________________________________||
signal.signal(signal.SIGINT, signal.SIG_IGN)

##__________________________________________________________________||
def main():

    try:
        run_tasks()
    except Exception as e:
        print_logs(e)
        raise

##__________________________________________________________________||
def run_tasks():

    setup()

    for package_path in args.paths:

        # e.g., package_path = 'task_00003.p.gz',
        # a relative to the directory in which this file is stored

        thisdir = os.path.dirname(__file__)
        package_path = os.path.join(thisdir, package_path)
        # e.g., package_path = 'c/d/task_00003.p.gz'

        result = run(package_path)

        result_path = compose_result_path(package_path)
        # e.g., '/a/b/c/d/results/task_00003/result.p.gz'

        store_result(result, result_path)

##__________________________________________________________________||
def print_logs(error):
    import socket
    import datetime
    messages = [
        '',
        '{:>20}: {}'.format('error', error),
        '{:>20}: {}'.format('sys.argv', sys.argv),
        '{:>20}: {}'.format('time', datetime.datetime.now()),
        '{:>20}: {}'.format('hostname', socket.getfqdn()),
        '{:>20}: {}'.format('pwd', os.getcwd()),
        '{:>20}: {}'.format('ls', ', '.join(sorted(os.listdir(os.getcwd())))),
        '{:>20}: {}'.format('sys.path', ', '.join(sys.path)),
        '{:>20}: {}'.format('os.environ', os.environ)

    ]
    print('\n'.join(messages), file=sys.stderr)

##__________________________________________________________________||
def setup():
    setup_logging()
    setup_python_modules()

def setup_logging():
    path = 'logging_levels.json.gz'
    if not os.path.isfile(path):
        return
    with gzip.GzipFile(path, 'r') as f:
        loglevel_dict = json.loads(f.read().decode('utf-8'))

    for name, level in loglevel_dict.items():
        logger = logging.getLogger(name)
        logger.setLevel(level)

    handler = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    handler.setFormatter(formatter)
    logging.getLogger('').addHandler(handler)


def setup_python_modules():

    dirname = 'python_modules'
    tarname = dirname + '.tar.gz'

    if os.path.exists(tarname) and not os.path.exists(dirname):
        if try_make_file('.untarring'):
            tar = tarfile.open(tarname)
            tar.extractall()
            tar.close()
            os.remove('.untarring')

    while os.path.isfile('.untarring'):
       pass

    if not os.path.exists(dirname): return

    sys.path.insert(0, dirname)

##_______________________________________________________________||
# http://stackoverflow.com/questions/33223564/atomically-creating-a-file-if-it-doesnt-exist-in-python
def try_make_file(filename):
    try:
        os.open(filename,  os.O_CREAT | os.O_EXCL)
        return True
    except OSError:
        # FileExistsError can be used for Python 3
        return False

##__________________________________________________________________||
def run(package_path):
    f = gzip.open(package_path, 'rb')
    package = pickle.load(f)
    result = package.task(*package.args, **package.kwargs)
    return result

##__________________________________________________________________||
def compose_result_path(package_path):

    # e.g., package_path = 'c/d/task_00003.p.gz'

    taskdir = os.path.dirname(os.path.abspath(package_path))
    # e.g., '/a/b/c/d'

    result_topdir = os.path.join(taskdir, 'results')
    # e.g., '/a/b/c/d/results'

    package_basename =  os.path.basename(package_path)
    # e.g., 'task_00003.p.gz'

    resultdir_basename = os.path.splitext(package_basename)[0]
    resultdir_basename = os.path.splitext(resultdir_basename)[0]
    # e.g., 'task_00003'

    resultdir = os.path.join(result_topdir, resultdir_basename)
    # e.g., '/a/b/c/d/results/task_00003'

    result_path = os.path.join(resultdir, 'result.p.gz')
    # e.g., '/a/b/c/d/results/task_00003/result.p.gz'

    return result_path

##__________________________________________________________________||
def store_result(result, result_path):
    mkdir_p(os.path.dirname(result_path))
    with gzip.open(result_path, 'wb') as f:
        pickle.dump(result, f, protocol=pickle.HIGHEST_PROTOCOL)

##__________________________________________________________________||
def mkdir_p(path):
    # http://stackoverflow.com/questions/600268/mkdir-p-functionality-in-python
    try:
        os.makedirs(path)
    except OSError as exc: # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else: raise

##__________________________________________________________________||
def print_profile_func(func, profile_out_path=None):
    result = profile_func(func)
    if profile_out_path is None:
        print(result)
    else:
        with open(profile_out_path, 'w') as f:
            f.write(result)

##__________________________________________________________________||
def profile_func(func):
    pr = cProfile.Profile()
    pr.enable()
    func()
    pr.disable()
    sortby = 'cumulative'
    try:
        s = StringIO()
        pstats.Stats(pr, stream=s).strip_dirs().sort_stats(sortby).print_stats()
    except TypeError:
        s = BytesIO()
        pstats.Stats(pr, stream=s).strip_dirs().sort_stats(sortby).print_stats()
    return s.getvalue()

##__________________________________________________________________||
if __name__ == '__main__':
    if args.profile:
        print_profile_func(
            func=main,
            profile_out_path=args.profile_out_path
        )
    else:
        main()

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import logging
import sys

import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

has_jupyter_notebook = False
try:
    from alphatwirl.progressbar import ProgressBarJupyter
    has_jupyter_notebook = True
except ImportError:
    pass

from alphatwirl.parallel import build_parallel
from alphatwirl.parallel.build import build_parallel_multiprocessing

##__________________________________________________________________||
@pytest.fixture(
    params=[True, False]
)
def isatty(request, monkeypatch):
    ret = request.param
    org_stdout = sys.stdout
    f = mock.Mock(**{
        'stdout.isatty.return_value': ret,
        'stdout.write.side_effect': lambda x : org_stdout.write(x)
    })
    module = sys.modules['alphatwirl.parallel.build']
    monkeypatch.setattr(module, 'sys', f)
    return ret

##__________________________________________________________________||
if has_jupyter_notebook:
    is_jupyter_notebook_parames = [True, False]
else:
    is_jupyter_notebook_parames = [False]

@pytest.fixture(params=is_jupyter_notebook_parames)
def is_jupyter_notebook(request, monkeypatch):
    ret = request.param
    f = mock.Mock()
    f.return_value = ret
    module = sys.modules['alphatwirl.parallel.build']
    monkeypatch.setattr(module, 'is_jupyter_notebook', f)
    return ret

##__________________________________________________________________||
@pytest.mark.parametrize('processes', [0, 1, 3])
@pytest.mark.parametrize('quiet', [True, False])
def test_build_parallel_multiprocessing(quiet, processes, isatty, is_jupyter_notebook):

    parallel_mode = 'multiprocessing'
    parallel = build_parallel(
        parallel_mode=parallel_mode,
        quiet=quiet,
        processes=processes,
    )

    ## progressMonitor
    if quiet:
        assert 'NullProgressMonitor' == parallel.progressMonitor.__class__.__name__
    elif processes == 0:
        assert 'ProgressMonitor' == parallel.progressMonitor.__class__.__name__
    else:
        assert 'BProgressMonitor' == parallel.progressMonitor.__class__.__name__

    if not quiet:
        if isatty:
            assert 'ProgressBar' == parallel.progressMonitor.presentation.__class__.__name__
        elif is_jupyter_notebook:
            assert 'ProgressBarJupyter' == parallel.progressMonitor.presentation.__class__.__name__
        else:
            assert 'ProgressPrint' == parallel.progressMonitor.presentation.__class__.__name__

    ## communicationChannel
    if processes == 0:
        assert 'CommunicationChannel0' == parallel.communicationChannel.__class__.__name__
    else:
        assert 'CommunicationChannel' == parallel.communicationChannel.__class__.__name__
        assert 'MultiprocessingDropbox' ==  parallel.communicationChannel.dropbox.__class__.__name__

    ## workingarea
    assert parallel.workingarea is None

##__________________________________________________________________||
def test_build_logging_unknown_parallel_mode(caplog):

    with caplog.at_level(logging.WARNING):
        parallel = build_parallel(parallel_mode='unknown_mode')

    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'WARNING'
    assert 'parallel.build' in caplog.records[0].name
    assert 'unknown parallel_mode' in caplog.records[0].msg

    assert 'MultiprocessingDropbox' ==  parallel.communicationChannel.dropbox.__class__.__name__

##__________________________________________________________________||
@pytest.mark.parametrize('dispatcher_options', [dict()])
@pytest.mark.parametrize('user_modules', [[], ['scribblers']])
def test_build_parallel_subprocess(user_modules, dispatcher_options):

    parallel_mode = 'subprocess'

    parallel = build_parallel(
        parallel_mode=parallel_mode,
        user_modules=user_modules,
        dispatcher_options=dispatcher_options
    )

    ## progressMonitor
    assert 'NullProgressMonitor' == parallel.progressMonitor.__class__.__name__

    ## communicationChannel
    assert 'CommunicationChannel' == parallel.communicationChannel.__class__.__name__
    assert 'TaskPackageDropbox' == parallel.communicationChannel.dropbox.__class__.__name__

    ## dispatcher
    assert 'SubprocessRunner' == parallel.communicationChannel.dropbox.dispatcher.__class__.__name__

    assert 'WorkingArea' == parallel.workingarea.__class__.__name__

    if user_modules:
        assert set(['scribblers', 'alphatwirl']) == set(parallel.workingarea.python_modules)
    else:
        assert set(['alphatwirl']) == set(parallel.workingarea.python_modules)


@pytest.mark.parametrize('dispatcher_options', [dict(), dict(job_desc_dict=dict(request_memory='120'))])
@pytest.mark.parametrize('htcondor_job_desc_extra', [[], ['request_memory = 250']])
@pytest.mark.parametrize('user_modules', [[], ['scribblers']])
def test_build_parallel_htcondor(
        user_modules, htcondor_job_desc_extra,
        dispatcher_options):

    parallel_mode = 'htcondor'

    parallel = build_parallel(
        parallel_mode=parallel_mode,
        user_modules=user_modules,
        htcondor_job_desc_extra=htcondor_job_desc_extra,
        dispatcher_options=dispatcher_options
    )

    ## progressMonitor
    assert 'NullProgressMonitor' == parallel.progressMonitor.__class__.__name__

    ## communicationChannel
    assert 'CommunicationChannel' == parallel.communicationChannel.__class__.__name__
    assert 'TaskPackageDropbox' == parallel.communicationChannel.dropbox.__class__.__name__

    ## dispatcher
    assert 'HTCondorJobSubmitter' == parallel.communicationChannel.dropbox.dispatcher.__class__.__name__
    assert htcondor_job_desc_extra == parallel.communicationChannel.dropbox.dispatcher.job_desc_extra
    if 'job_desc_dict' in dispatcher_options:
        assert dispatcher_options['job_desc_dict'] == parallel.communicationChannel.dropbox.dispatcher.user_job_desc_dict

    assert 'WorkingArea' == parallel.workingarea.__class__.__name__

    if user_modules:
        assert set(['scribblers', 'alphatwirl']) == set(parallel.workingarea.python_modules)
    else:
        assert set(['alphatwirl']) == set(parallel.workingarea.python_modules)


##__________________________________________________________________||

##__________________________________________________________________||
def test_build_depricated(caplog):
    with caplog.at_level(logging.WARNING):
        parallel = build_parallel_multiprocessing(quiet=True, processes=4)

    progressMonitor = parallel.progressMonitor
    communicationChannel = parallel.communicationChannel
    assert 'NullProgressMonitor' == progressMonitor.__class__.__name__
    assert 'CommunicationChannel' == communicationChannel.__class__.__name__
    assert 'MultiprocessingDropbox' == communicationChannel.dropbox.__class__.__name__

    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'WARNING'
    assert 'parallel.build' in caplog.records[0].name
    assert 'deprecated' in caplog.records[0].msg

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

from alphatwirl.selection.factories.factory import AllFactory
from alphatwirl.selection.factories.factory import AnyFactory
from alphatwirl.selection.factories.factory import NotFactory

##__________________________________________________________________||
def test_AllFactory():
    MockClass = mock.Mock()
    component1 = mock.sentinel.component1
    component2 = mock.sentinel.component1
    components = (component1, component2)
    name = mock.sentinel.name
    obj = AllFactory(components=components, AllClass=MockClass, name=name)
    assert [mock.call(name=name)] == MockClass.call_args_list
    assert [
        mock.call(component1),
        mock.call(component2),
    ] == MockClass().add.call_args_list

def test_AnyFactory():
    MockClass = mock.Mock()
    component1 = mock.sentinel.component1
    component2 = mock.sentinel.component1
    components = (component1, component2)
    name = mock.sentinel.name
    obj = AnyFactory(components=components, AnyClass=MockClass, name=name)
    assert [mock.call(name=name)] == MockClass.call_args_list
    assert [
        mock.call(component1),
        mock.call(component2),
    ] == MockClass().add.call_args_list

def test_NotFactory():
    MockClass = mock.Mock()
    component1 = mock.sentinel.component1
    components = (component1, )
    name = mock.sentinel.name
    obj = NotFactory(components=components, NotClass=MockClass, name=name)
    assert [mock.call(selection=component1, name=name)] == MockClass.call_args_list
    assert obj == MockClass()


##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import array

import ROOT

from .inspect import is_ROOT_null_pointer

##__________________________________________________________________||
class BranchAddressManager(object):
    """The branch address manager for ROOT TTree

    This class manages array.array objects used for branch addresses
    of ROOT TTree. The main purpose of this class is to prevent
    multiple arrays from being created for the same branch.


    All instances of this class share the data.
    """

    arrayDict = { }
    counterArrayDict = { }

    def getArrays(self, tree, branchName):
        """return the array.array objects for the branch and its counter branch

        This method returns a pair of the array.array objects. The first one is
        for the given tree and branch name. The second one is for its counter
        branch. The second one will be None when the branch does not have a
        counter. A pair of None will be returned when the tree does not have
        the branch.

        """
        itsArray = self._getArray(tree, branchName)
        if itsArray is None: return None, None
        itsCountArray = self._getCounterArray(tree, branchName)
        return itsArray, itsCountArray

    def _getArray(self, tree, branchName):

        if (tree, branchName) in self.__class__.arrayDict:
            return self.__class__.arrayDict[(tree, branchName)]

        self._createArraysForBranchAndCounter(tree, branchName)
        return self.__class__.arrayDict[(tree, branchName)]

    def _getCounterArray(self, tree, branchName):
        if (tree, branchName) in self.__class__.counterArrayDict:
            return self.__class__.counterArrayDict[(tree, branchName)]
        return None

    def _createArraysForBranchAndCounter(self, tree, branchName):

        leafNames = [l.GetName() for l in tree.GetListOfLeaves()]
        if branchName not in leafNames:
            self.__class__.arrayDict[(tree, branchName)] = None
            return

        leafInfo = inspectLeaf(tree, branchName)
        if leafInfo['arraytype'] is None:
            self.__class__.arrayDict[(tree, branchName)] = None
            return

        tree.SetBranchStatus(leafInfo['name'], 1)
        if leafInfo['countname'] is not None: tree.SetBranchStatus(leafInfo['countname'], 1)

        maxn = 1 if leafInfo['countmax'] is None or leafInfo['countmax'] == 0 else leafInfo['countmax']
        itsArray = array.array(leafInfo['arraytype'], maxn*[ 0 ])
        tree.SetBranchAddress(leafInfo['name'], itsArray)

        if leafInfo['countname'] is not None:
            itsCountArray = self._getArray(tree, leafInfo['countname'])
        else:
            itsCountArray = None

        self.__class__.arrayDict[(tree, branchName)] = itsArray
        self.__class__.counterArrayDict[(tree, branchName)] = itsCountArray

##__________________________________________________________________||
def inspectLeaf(tree, bname):


    # This dict maps a ROOT type to a type code of the python array
    # https://root.cern.ch/root/html/RtypesCore.h
    # https://docs.python.org/2/library/array.html
    typedic = dict(
        Char_t='b',
        UChar_t='B',
        Short_t='h',
        UShort_t='H',
        Int_t='i',
        UInt_t='I',
        Float_t='f',
        Double_t='d',
        Long64_t='l',
        ULong64_t='L',
        Bool_t='b',
    )

    leaf = tree.GetLeaf(bname)
    leafcount = leaf.GetLeafCount()
    isArray = not is_ROOT_null_pointer(leafcount)

    countmax = None
    if isArray:
        countmax = _get_countmax(leafcount, tree, bname)

    return dict(
        name=leaf.GetName(),
        ROOTtype=leaf.GetTypeName(),
        arraytype=typedic[leaf.GetTypeName()] if leaf.GetTypeName() in typedic else None,
        isarray=isArray,
        countname=leafcount.GetName() if isArray else None,
        countROOTtype=leafcount.GetTypeName() if isArray else None,
        countarraytype=typedic[leafcount.GetTypeName()] if isArray else None,
        countmax=countmax,
        )

def _get_countmax(leafcount, tree, bname):
    # If the tree is a chain, leafcount.GetMaximum() only returns the
    # maximum in the current file. The `countmax` needs to be the
    # maximum in all files. Not very efficient or elegant, the current
    # implementation opens all files in the chain and finds the
    # maximum.

    try:
        tobjarray_files = tree.GetListOfFiles()
    except AttributeError:
        # the tree is not a chain
        return leafcount.GetMaximum()

    if 0 == tobjarray_files.GetLast():
        # the chain has only one file
        return leafcount.GetMaximum()

    filepahts = (f.GetTitle() for f in tobjarray_files)
    files = (ROOT.TFile.Open(p) for p in filepahts)
    trees = (f.Get(tree.GetName()) for f in files)
    leaves = (t.GetLeaf(bname) for t in trees)
    leafcounts = (l.GetLeafCount() for l in leaves)
    countmaxs = (l.GetMaximum() for l in leafcounts)
    countmax = max(countmaxs)
    return countmax

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>

from .ReturnTrue import ReturnTrue

##__________________________________________________________________||
class Binning(object):
    def __init__(self, boundaries=None, lows=None, ups=None,
                 retvalue='lowedge', bins=None, underflow_bin=None, overflow_bin=None,
                 valid=ReturnTrue()):

        if boundaries is None:
            if lows is None or ups is None:
                raise ValueError("Only either boundaries or pairs of lows and ups need to be given!")
            if not tuple(lows[1:]) == tuple(ups[:-1]):
                raise ValueError("Boundaries cannot be determined from lows=" + str(lows) + " and ups=" + str(ups))
            self.boundaries = tuple(lows) + (ups[-1], )
            self.lows = tuple(lows)
            self.ups = tuple(ups)
        else:
            if lows is not None or ups is not None:
                raise ValueError("Only either boundaries or pairs of lows and ups need to be given!")
            if len(boundaries) < 2:
                raise ValueError("Needs at least one bin! boundaries=" + str(boundaries))
            self.boundaries = tuple(boundaries)
            self.lows = tuple(boundaries[:-1])
            self.ups = tuple(boundaries[1:])

        supportedRetvalues = ('number', 'lowedge')
        if retvalue not in supportedRetvalues:
            raise ValueError("The retvalue '%s' is not supported! " % (retvalue, ) + "Supported values are '" + "', '".join(supportedRetvalues)  + "'")

        self.lowedge = (retvalue == 'lowedge')
        if self.lowedge:
            if bins is not None: raise ValueError("bins cannot be given when retvalue is '" + retvalue + "'!")
            if underflow_bin is not None: raise ValueError("underflow_bin cannot be given when retvalue is '" + retvalue + "'!")
            if overflow_bin is not None: raise ValueError("overflow_bin cannot be given when retvalue is '" + retvalue + "'!")

        if self.lowedge:
            self.bins = self.lows
            self.underflow_bin = float("-inf")
            self.overflow_bin = self.ups[-1]
        else:
            self.bins = bins if bins is not None else tuple(range(1, len(self.lows) + 1))
            self.underflow_bin = underflow_bin if underflow_bin is not None else min(self.bins) - 1
            self.overflow_bin = overflow_bin if overflow_bin is not None else max(self.bins) + 1

        self._valid = valid

    def __repr__(self):
        return '{}(boundaries={!r}, underflow_bin={!r}, overflow_bin={!r}, valid={!r})'.format(
            self.__class__.__name__,
            self.boundaries, self.underflow_bin, self.overflow_bin, self._valid
        )

    def __call__(self, val):
        if not self._valid(val): return None
        if val < self.lows[0]: return self.underflow_bin
        if self.ups[-1] <= val: return self.overflow_bin
        return [b for b, l, u in zip(self.bins, self.lows, self.ups) if l <= val < u][0]

    def next(self, bin):
        if self.lowedge:
            # call self._call__() to ensure that the 'bin' is indeed one of the
            # bins.
            bin = self.__call__(bin)

        if bin == self.underflow_bin: return self.bins[0]
        if bin == self.overflow_bin: return self.overflow_bin
        if bin == self.bins[-1]: return self.overflow_bin
        return self.bins[self.bins.index(bin) + 1]

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import pytest
import time
import multiprocessing

try:
    import unittest.mock as mock
except ImportError:
    import mock

from alphatwirl.progressbar import ProgressReportPickup

##__________________________________________________________________||
@pytest.fixture()
def presentation():
    return mock.MagicMock()

##__________________________________________________________________||
@pytest.fixture()
def queue():
    return multiprocessing.Queue()

@pytest.fixture()
def pickup(queue, presentation):
    return ProgressReportPickup(queue, presentation)

##__________________________________________________________________||
def test_start_join(pickup, queue, presentation):
    presentation.nreports.return_value = 10
    pickup.start()
    queue.put(None)
    pickup.join()

##__________________________________________________________________||
@pytest.fixture()
def mock_queue():
    return mock.MagicMock()

@pytest.fixture()
def pickup0(mock_queue, presentation):
    return ProgressReportPickup(mock_queue, presentation)

##__________________________________________________________________||
def test_run_until_the_end_order_arrives_no_report(pickup0, mock_queue, presentation):

    mock_queue.empty.side_effect = [False, True]
    mock_queue.get.side_effect = [None]
    pickup0._run_until_the_end_order_arrives()

    assert [] == presentation.mock_calls

def test_run_until_the_end_order_arrives_one_report(pickup0, mock_queue, presentation):

    report = mock.MagicMock()
    mock_queue.empty.side_effect = [False, False, True]
    mock_queue.get.side_effect = [report, None]
    pickup0._run_until_the_end_order_arrives()

    assert [mock.call.present(report)] == presentation.mock_calls

def test_run_until_the_end_order_arrives_one_report_once_empty(pickup0, mock_queue, presentation):

    report1 = mock.MagicMock()
    mock_queue.empty.side_effect = [False, True, False, True] # it becomes empty once
    mock_queue.get.side_effect = [report1, None]
    pickup0._run_until_the_end_order_arrives()

    assert [mock.call.present(report1)] == presentation.mock_calls

def test_run_until_the_end_order_arrives_two_reports(pickup0, mock_queue, presentation):

    report1 = mock.MagicMock()
    report2 = mock.MagicMock()
    mock_queue.empty.side_effect = [False, False, False, True]
    mock_queue.get.side_effect = [report1, None, report2] # report2 arrives
                                                          # after the end_order
    pickup0._run_until_the_end_order_arrives()

    assert [mock.call.present(report1), mock.call.present(report2)] == presentation.mock_calls

##__________________________________________________________________||
@pytest.fixture()
def mocktime(monkeypatch):
    ret = mock.MagicMock(return_value = 1000.0)
    monkeypatch.setattr(time, 'time', ret)
    return ret

def test_run_until_reports_stop_coming_no_report(pickup0, mock_queue, presentation, mocktime):
    presentation.nreports.side_effect = [0]
    pickup0._run_until_reports_stop_coming()
    assert [] == presentation.present.mock_calls

def test_run_until_reports_stop_coming_one_report(pickup0, mock_queue, presentation, mocktime):
    presentation.nreports.side_effect = [1, 0]
    report = mock.MagicMock()
    mock_queue.empty.side_effect = [False, False, True]
    mock_queue.get.side_effect = [report, None]
    pickup0._run_until_reports_stop_coming()
    assert [mock.call(report)] == presentation.present.mock_calls

def test_run_until_reports_stop_coming_one_report_timeout(pickup0, mock_queue, presentation, mocktime):
    presentation.nreports.return_value = 1
    report = mock.MagicMock()
    mock_queue.empty.return_value = True
    mock_queue.get.side_effect = [report, None]
    mocktime.side_effect = [1000.0, 1003.0]
    pickup0._run_until_reports_stop_coming()
    assert [ ] == presentation.present.mock_calls

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
from __future__ import print_function
import pytest

import alphatwirl
from alphatwirl.progressbar import atpbar

try:
    import unittest.mock as mock
except ImportError:
    import mock

##__________________________________________________________________||
@pytest.fixture()
def mock_report_progress(monkeypatch):
    ret = mock.Mock()
    monkeypatch.setattr(alphatwirl.progressbar, 'report_progress', ret)
    return ret

##__________________________________________________________________||
class Iter(object):
    def __init__(self, content):
        self.content = content

    def __len__(self):
        return len(self.content)

    def __iter__(self):
        for e in self.content:
            yield e

class GetItem(object):
    def __init__(self, content):
        self.content = content

    def __len__(self):
        return len(self.content)

    def __getitem__(self, i):
        return self.content[i]

##__________________________________________________________________||
iterable_classes = [list, Iter, GetItem]

empty = [ ]
one = [mock.sentinel.item1]
three = [mock.sentinel.item1, mock.sentinel.item2, mock.sentinel.item3]
contents = [empty, one, three]
contents_ids = ['empty', 'one', 'three']

##__________________________________________________________________||
@pytest.mark.parametrize('content', contents, ids=contents_ids)
@pytest.mark.parametrize('iterable_class', iterable_classes)
def test_iterable(iterable_class, content):
    iterable = iterable_class(content)
    assert content == [e for e in iterable]

##__________________________________________________________________||
@pytest.mark.parametrize('content', contents, ids=contents_ids)
@pytest.mark.parametrize('iterable_class', iterable_classes)
def test_atpbar_iterables(mock_report_progress, iterable_class, content):
    iterable = iterable_class(content)

    ##
    returned = [ ]
    for e in atpbar(iterable):
        returned.append(e)

    ##
    assert content == returned

    ##
    assert len(content) + 1 == len(mock_report_progress.call_args_list)
    for i, c in enumerate(mock_report_progress.call_args_list):
        args, kwargs = c
        report = args[0]
        assert i == report.done
        assert len(content) == report.total

##__________________________________________________________________||
@pytest.mark.parametrize('content', contents, ids=contents_ids)
@pytest.mark.parametrize('iterable_class', iterable_classes)
def test_atpbar_enumerate(mock_report_progress, iterable_class, content):
    iterable = iterable_class(content)

    ##
    returned = [ ]
    for i, e in enumerate(atpbar(iterable)):
        returned.append(e)

    ##
    assert content == returned

    ##
    assert len(content) + 1 == len(mock_report_progress.call_args_list)
    for i, c in enumerate(mock_report_progress.call_args_list):
        args, kwargs = c
        report = args[0]
        assert i == report.done
        assert len(content) == report.total


##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import os
import time
import collections

import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

from alphatwirl.concurrently import MultiprocessingDropbox
from alphatwirl.concurrently import TaskPackage
from alphatwirl import progressbar

##__________________________________________________________________||
def test_init_raise():
    with pytest.raises(ValueError):
        MultiprocessingDropbox(nprocesses=0)

def test_open_close():
    obj = MultiprocessingDropbox()
    obj.open()
    obj.close()

def test_open_open_close():
    obj = MultiprocessingDropbox()
    obj.open()
    obj.open() # don't create workers again
    obj.close()

def test_repr():
    obj = MultiprocessingDropbox()
    repr(obj)

##__________________________________________________________________||
MockResult = collections.namedtuple('MockResult', 'name args kwargs')

class MockTask(object):
    def __init__(self, name, time):
        self.name = name
        self.time = time

    def __call__(self, *args, **kwargs):
        time.sleep(self.time)
        assert isinstance(progressbar._progress_reporter, MockProgressReporter)
        return MockResult(name=self.name, args=args, kwargs=kwargs)

##__________________________________________________________________||
@pytest.fixture()
def package1():
    task = MockTask(name='task1', time=0.010)
    args = (111, 222)
    kwargs = dict(A='abc', B='def')
    return TaskPackage(task=task, args=args, kwargs=kwargs)

@pytest.fixture()
def package2():
    task = MockTask(name='task2', time=0.001)
    args = ( )
    kwargs = { }
    return TaskPackage(task=task, args=args, kwargs=kwargs)

@pytest.fixture()
def package3():
    task = MockTask(name='task3', time=0.005)
    args = (33, 44)
    kwargs = { }
    return TaskPackage(task=task, args=args, kwargs=kwargs)

@pytest.fixture()
def package4():
    task = MockTask(name='task4', time=0.002)
    args = ( )
    kwargs = dict(ABC='abc', DEF='def')
    return TaskPackage(task=task, args=args, kwargs=kwargs)

##__________________________________________________________________||
class MockProgressReporter(object):
    pass

@pytest.fixture()
def mock_progressmonitor():
    ret = mock.MagicMock()
    ret.createReporter.return_value = MockProgressReporter()
    return ret

@pytest.fixture()
def obj(mock_progressmonitor):
    ret = MultiprocessingDropbox(progressMonitor=mock_progressmonitor)
    ret.open()
    yield ret
    ret.terminate()
    ret.close()

##__________________________________________________________________||
def test_put(obj, package1, package2):
    assert 0 == obj.put(package1)
    assert 1 == obj.put(package2)

def test_put_multiple(obj, package1, package2):
    assert [0, 1] == obj.put_multiple([package1, package2])

def test_put_receive(obj, package1, package2):
    packages = [package1, package2]
    pkgidxs = [ ]
    for p in packages:
        pkgidxs.append(obj.put(p))

    expected = [
        (i, MockResult(name=p.task.name, args=p.args, kwargs=p.kwargs))
        for i, p in zip(pkgidxs, packages)]
    actual = obj.receive()
    assert expected == actual

def test_receive_order(obj, package1, package2, package3):
    # results of tasks are sorted in the order in which the tasks are put.

    packages = [package1, package2, package3]
    pkgidxs = [ ]
    for p in packages:
        pkgidxs.append(obj.put(p))

    expected = [
        (i, MockResult(name=p.task.name, args=p.args, kwargs=p.kwargs))
        for i, p in zip(pkgidxs, packages)]
    actual = obj.receive()
    assert expected == actual

def test_put_receive_repeat(obj, package1, package2, package3, package4):

    packages = [package1, package2]
    pkgidxs = [ ]
    for p in packages:
        pkgidxs.append(obj.put(p))

    expected = [
        (i, MockResult(name=p.task.name, args=p.args, kwargs=p.kwargs))
        for i, p in zip(pkgidxs, packages)]
    actual = obj.receive()
    assert expected == actual

    packages = [package3, package4]
    pkgidxs = [ ]
    for p in packages:
        pkgidxs.append(obj.put(p))

    expected = [
        (i, MockResult(name=p.task.name, args=p.args, kwargs=p.kwargs))
        for i, p in zip(pkgidxs, packages)]
    actual = obj.receive()
    assert expected == actual

def test_begin_put_recive_end_repeat(obj, package1, package2):

    obj.put(package1)
    obj.receive()
    obj.close()
    obj.open()
    obj.put(package2)
    obj.receive()

def test_terminate(obj, package1, package2):

    obj.put(package1)
    obj.put(package2)
    obj.terminate()

def test_terminate_close(obj, package1, package2):

    obj.put(package1)
    obj.put(package2)
    obj.terminate()
    obj.close()

def test_receive_without_put(obj):
    assert [ ] == obj.receive()

##__________________________________________________________________||
def test_poll(obj, package1, package2, package3, package4):
    packages = [package1, package2, package3, package4]
    pkgidxs = obj.put_multiple(packages)

    expected = [
        (i, MockResult(name=p.task.name, args=p.args, kwargs=p.kwargs))
        for i, p in zip(pkgidxs, packages)]

    actual = [ ]
    while len(actual) < len(expected):
        actual.extend(obj.poll())
    assert sorted(expected) == sorted(actual)

def test_poll_then_receive(obj, package1, package2, package3, package4):
    packages = [package1, package2, package3, package4]
    pkgidxs = obj.put_multiple(packages)

    expected = [
        (i, MockResult(name=p.task.name, args=p.args, kwargs=p.kwargs))
        for i, p in zip(pkgidxs, packages)]

    actual = [ ]
    while not actual:
        actual.extend(obj.poll())

    actual.extend(obj.receive())

    assert sorted(expected) == sorted(actual)

##__________________________________________________________________||
def test_receive_one(obj, package1, package2, package3, package4):
    packages = [package1, package2, package3, package4]
    pkgidxs = obj.put_multiple(packages)

    expected = [
        (i, MockResult(name=p.task.name, args=p.args, kwargs=p.kwargs))
        for i, p in zip(pkgidxs, packages)]

    actual = [ ]
    actual.append(obj.receive_one())
    actual.append(obj.receive_one())
    actual.append(obj.receive_one())
    actual.append(obj.receive_one())
    assert obj.receive_one() is None
    assert sorted(expected) == sorted(actual)

def test_receive_one_then_receive(obj, package1, package2, package3, package4):
    packages = [package1, package2, package3, package4]
    pkgidxs = obj.put_multiple(packages)

    expected = [
        (i, MockResult(name=p.task.name, args=p.args, kwargs=p.kwargs))
        for i, p in zip(pkgidxs, packages)]

    time.sleep(0.005) # so multiple jobs finish

    actual = [ ]
    actual.append(obj.receive_one())

    actual.extend(obj.receive())

    assert sorted(expected) == sorted(actual)

def test_receive_one_then_poll(obj, package1, package2, package3, package4):
    packages = [package1, package2, package3, package4]
    pkgidxs = obj.put_multiple(packages)

    expected = [
        (i, MockResult(name=p.task.name, args=p.args, kwargs=p.kwargs))
        for i, p in zip(pkgidxs, packages)]

    time.sleep(0.005) # so multiple jobs finish

    actual = [ ]
    actual.append(obj.receive_one())

    while len(actual) < len(expected):
        actual.extend(obj.poll())

    assert sorted(expected) == sorted(actual)

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
from .BEvents import BEvents
from .EventBuilder import EventBuilder

from alphatwirl.misc.deprecation import _deprecated

##__________________________________________________________________||
@_deprecated(msg='use BuildEvents instead')
class BEventBuilder(EventBuilder):
    def __init__(self, config):
        super(BEventBuilder, self).__init__(config, EventsClass=BEvents)

##__________________________________________________________________||



import unittest

from .mock import MockSummarizer

##__________________________________________________________________||
class TestMockSummarizer(unittest.TestCase):

    def setUp(self):
        self.obj1 = MockSummarizer((1, 2))
        self.obj2 = MockSummarizer((3, 4))

    def tearDown(self):
        pass

    def test_repr(self):
        repr(self.obj1)

    def test_add(self):
        obj3 = self.obj1 + self.obj2
        self.assertEqual((1, 2, 3, 4), obj3._results)

    def test_radd(self):
        obj3 = sum([self.obj1, self.obj2])
        self.assertEqual((1, 2, 3, 4), obj3._results)

    def test_to_tuple_list(self):
        self.assertEqual((1, 2), self.obj1.to_tuple_list())

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import logging
import pytest

import alphatwirl
from alphatwirl.progressbar import atpbar

try:
    import unittest.mock as mock
except ImportError:
    import mock

##__________________________________________________________________||
@pytest.fixture()
def mock_report_progress(monkeypatch):
    ret = mock.Mock()
    monkeypatch.setattr(alphatwirl.progressbar, 'report_progress', ret)
    return ret

@pytest.fixture()
def mock_report_progress_raise(monkeypatch):
    ret = mock.Mock()
    ret.side_effect = Exception
    monkeypatch.setattr(alphatwirl.progressbar, 'report_progress', ret)
    return ret

##__________________________________________________________________||
class Iter(object):
    def __init__(self, content):
        self.content = content

    def __repr__(self):
        return self.__class__.__name__

    def __len__(self):
        return len(self.content)

    def __iter__(self):
        for e in self.content:
            yield e

##__________________________________________________________________||
content = [mock.sentinel.item1, mock.sentinel.item2, mock.sentinel.item3]

##__________________________________________________________________||
def test_atpbar_name_repr(mock_report_progress, caplog):

    iterable = Iter(content)
    returned = [e for e in atpbar(iterable)]

    ##
    assert content == returned

    ##
    assert len(content) + 1 == len(mock_report_progress.call_args_list)
    for i, c in enumerate(mock_report_progress.call_args_list):
        args, kwargs = c
        report = args[0]
        assert i == report.done
        assert len(content) == report.total
        assert 'Iter' == report.name # repr(iterable)
        print(report)

##__________________________________________________________________||
def test_atpbar_name_given(mock_report_progress, caplog):

    iterable = Iter(content)
    returned = [e for e in atpbar(iterable, name='given')]

    ##
    assert content == returned

    ##
    assert len(content) + 1 == len(mock_report_progress.call_args_list)
    for i, c in enumerate(mock_report_progress.call_args_list):
        args, kwargs = c
        report = args[0]
        assert i == report.done
        assert len(content) == report.total
        assert 'given' == report.name
        print(report)

##__________________________________________________________________||
def test_atpbar_raise(mock_report_progress_raise, caplog):

    iterable = Iter(content)
    returned = [e for e in atpbar(iterable, name='given')]

    ##
    assert content == returned

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import os
from .TableFileNameComposer import TableFileNameComposer
from ..summary import Count, WeightCalculatorOne

##__________________________________________________________________||
class TableConfigCompleter(object):
    """
    an example complete config::

        tblcfg = {
            'keyAttrNames': ('met_pt',),
            'binnings': (MockBinning(),),
            'keyIndices': None
            'valAttrNames' : None,
            'valIndices' : None,
            'keyOutColumnNames': ('met_pt',),
            'valOutColumnNames': ('n', 'nvar'),
            'weight': MockWeight(),
            'summaryClass': Count,
            'sort': True,
            'outFile': True,
            'outFileName': 'tbl_n_component_met_pt.txt',
            'outFilePath': '/tmp/tbl_n_component_met_pt.txt',
        }

    """
    def __init__(self,
                 defaultSummaryClass=Count,
                 defaultWeight=WeightCalculatorOne(),
                 defaultOutDir='.',
                 createOutFileName=TableFileNameComposer()):

        self.defaultSummaryClass = defaultSummaryClass
        self.defaultWeight = defaultWeight
        self.defaultOutDir = defaultOutDir
        self.createOutFileName = createOutFileName

        self.default_cfg = dict(
            keyAttrNames=( ),
            binnings=None,
            keyIndices=None,
            valAttrNames=None,
            valIndices=None,
            summaryClass=defaultSummaryClass,
            weight=defaultWeight,
            outFile=True,
            sort=True,
            nevents=None,
        )

    def __repr__(self):
        name_value_pairs = (
            ('defaultSummaryClass', self.defaultSummaryClass),
            ('defaultWeight', self.defaultWeight),
            ('defaultOutDir', self.defaultOutDir),
            ('createOutFileName', self.createOutFileName),
        )
        return '{}({})'.format(
            self.__class__.__name__,
            ', '.join(['{}={!r}'.format(n, v) for n, v in name_value_pairs]),
        )

    def complete(self, tblcfg):
        ret = self.default_cfg.copy()
        ret.update(tblcfg)

        use_defaultSummaryClass = 'summaryClass' not in tblcfg

        ret['keyOutColumnNames'] = ret.get('keyOutColumnNames', ret['keyAttrNames'])
        # TODO: this line is not tested well. The following code also passes the tests
        # ret['keyOutColumnNames'] = ret.get('keyAttrNames', ret['keyAttrNames'])

        if 'valOutColumnNames' not in ret:
            if use_defaultSummaryClass:
                ret['valOutColumnNames'] = ('n', 'nvar')
            else:
                ret['valOutColumnNames'] = ret['valAttrNames'] if ret['valAttrNames'] is not None else ()

        if ret['outFile']:
            if 'outFileName' not in ret:
                if use_defaultSummaryClass:
                    ret['outFileName'] = self.createOutFileName(
                        ret['keyOutColumnNames'], ret['keyIndices']
                    )
                else:
                    keyOutColumnNames  = ret['keyOutColumnNames'] if ret['keyOutColumnNames'] is not None else ()
                    keyIndices = ret['keyIndices'] if ret['keyIndices'] is not None else (None, )*len(keyOutColumnNames)
                    valOutColumnNames  = ret['valOutColumnNames'] if ret['valOutColumnNames'] is not None else ()
                    valIndices = ret['valIndices'] if ret['valIndices'] is not None else (None, )*len(valOutColumnNames)
                    ret['outFileName'] = self.createOutFileName(
                        keyOutColumnNames + valOutColumnNames,
                        keyIndices + valIndices,
                        prefix='tbl_{}'.format(ret['summaryClass'].__name__)
                    )
            if 'outFilePath' not in ret:
                ret['outFilePath'] = os.path.join(self.defaultOutDir, ret['outFileName'])
        return ret

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import threading
import time

##__________________________________________________________________||
class ProgressReportPickup(threading.Thread):
    def __init__(self, queue, presentation):
        threading.Thread.__init__(self)
        self.queue = queue
        self.presentation = presentation
        self.last_wait_time = 1.0 # [second]

    def run(self):
        try:
            self._run_until_the_end_order_arrives()
            self._run_until_reports_stop_coming()
        except KeyboardInterrupt:
            pass

    def _run_until_the_end_order_arrives(self):
        end_order_arrived = False
        while True:
            while not self.queue.empty():
                report = self.queue.get()
                if report is None: # the end order
                    end_order_arrived = True
                    continue
                self.presentation.present(report)
            if end_order_arrived: break

    def _run_until_reports_stop_coming(self):
        self._read_time()
        while self.presentation.nreports() > 0:
            if self._time() - self.last_time > self.last_wait_time: break
            while not self.queue.empty():
                report = self.queue.get()
                if report is None: continue
                self._read_time()
                self.presentation.present(report)

    def _time(self): return time.time()
    def _read_time(self): self.last_time = self._time()

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>

from alphatwirl.progressbar import atpbar

##__________________________________________________________________||
class ReaderComposite(object):

    """A composite of event readers"

    This class is a composite in the composite pattern.

    Examples of event readers are instances of `Summarizer` and this
    class.

    When `event()` is called, it calls `event()` of each reader in the
    order in which the readers are added. If a reader returns `False`,
    it won't call the remaining readers.

    """

    def __init__(self, readers=None):
        if readers is None:
            readers = [ ]
        self.readers = list(readers)

    def __repr__(self):
        return '{}({!r})'.format(
            self.__class__.__name__,
            self.readers
        )

    def add(self, reader):
        self.readers.append(reader)

    def begin(self, event):
        for reader in self.readers:
            if not hasattr(reader, 'begin'):
                continue
            reader.begin(event)

    def event(self, event):
        for reader in self.readers:
            if reader.event(event) is False:
                break

    def end(self):
        for reader in self.readers:
            if not hasattr(reader, 'end'):
                continue
            reader.end()

    def merge(self, other):
        for r, o in zip(self.readers, other.readers):
            if not hasattr(r, 'merge'):
                continue
            r.merge(o)

    def collect(self):
        ret = [ ]
        for reader in atpbar(self.readers, name='collecting results'):
            if not hasattr(reader, 'collect'):
                ret.append(None)
                continue
            ret.append(reader.collect())
        return ret

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import sys
import logging
import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

has_no_ROOT = False
try:
    import ROOT
except ImportError:
    has_no_ROOT = True


if not has_no_ROOT:
    from alphatwirl.roottree import BuildEvents

##__________________________________________________________________||
pytestmark = pytest.mark.skipif(has_no_ROOT, reason="has no ROOT")

##__________________________________________________________________||
@pytest.fixture()
def mockroot(monkeypatch):
    ret = mock.Mock()
    module = sys.modules['alphatwirl.roottree.build']
    monkeypatch.setattr(module, 'ROOT', ret)
    return ret

@pytest.fixture()
def mockevents(monkeypatch):
    ret = mock.Mock()
    return ret

@pytest.fixture()
def mocktfile():
    ret = mock.Mock()
    ret.good = True
    ret.IsZombie.return_value = False
    return ret

@pytest.fixture()
def mocktfile_null():
    ret = mock.Mock()
    ret.good = False
    ret.GetName.side_effect = ReferenceError
    return ret

@pytest.fixture()
def mocktfile_zombie():
    ret = mock.Mock()
    ret.good = False
    ret.IsZombie.return_value = True
    return ret

@pytest.fixture(
    params=[
        ['good', 'good'],
        ['null', 'good'],
        ['zombie', 'good'],
    ]
)
def files(request, mocktfile, mocktfile_null, mocktfile_zombie):
    map_ = dict(good=mocktfile, null=mocktfile_null, zombie=mocktfile_zombie)
    ret = [map_[p] for p in request.param]
    return ret

@pytest.fixture()
def config(mockevents):
    return dict(
        events_class=mockevents,
        file_paths=['/path/to/input1/tree.root', '/path/to/input2/tree.root'],
        tree_name='tree',
        max_events=123,
        start=11,
        check_files=True,
        skip_error_files=True
    )

@pytest.fixture()
def obj(config, mockroot):
    return BuildEvents(config)

##__________________________________________________________________||
def test_repr(obj):
    repr(obj)

def test_build(config, files, mockroot, mockevents):
    config['check_files'] = False
    obj = BuildEvents(config)
    mockroot.TFile.Open.side_effect = files
    events = obj()
    assert [mock.call('tree')] == mockroot.TChain.call_args_list
    chain = mockroot.TChain()
    assert [
        mock.call('/path/to/input1/tree.root'),
        mock.call('/path/to/input2/tree.root'),
    ] == chain.Add.call_args_list
    assert [mock.call(chain, 123, 11)] == mockevents.call_args_list

def test_build_skip(config, files, mockroot, mockevents, caplog):
    config['check_files'] = True
    config['skip_error_files'] = True
    obj = BuildEvents(config)
    mockroot.TFile.Open.side_effect = files
    with caplog.at_level(logging.WARNING):
        events = obj()
    assert [mock.call('tree')] == mockroot.TChain.call_args_list
    chain = mockroot.TChain()
    file_paths = [p for p, f in zip(config['file_paths'], files) if f.good]
    assert [mock.call(p) for p in file_paths] == chain.Add.call_args_list
    assert [mock.call(chain, 123, 11)] == mockevents.call_args_list

    assert len(caplog.records) == sum([not f.good for f in files])

    if all([f.good for f in files]):
        return

    assert caplog.records[0].levelname == 'WARNING'
    assert 'cannot open' in caplog.records[0].msg

def test_build_raise(config, files, mockroot, caplog):
    config['check_files'] = True
    config['skip_error_files'] = False
    obj = BuildEvents(config)
    mockroot.TFile.Open.side_effect = files
    if all([f.good for f in files]):
        return

    with caplog.at_level(logging.WARNING):
        with pytest.raises(OSError):
            events = obj()

    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'ERROR'
    assert 'cannot open' in caplog.records[0].msg

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import uuid
import logging

import alphatwirl
from .ProgressReport import ProgressReport

##__________________________________________________________________||
def atpbar(iterable, name=None):
    """Progress bar

    """
    try:
        len_ = len(iterable)
    except TypeError:
        logger = logging.getLogger(__name__)
        logging.warning('length is unknown: {!r}'.format(iterable))
        logging.warning('atpbar is turned off')
        return iterable

    if name is None:
        name = repr(iterable)

    return Atpbar(iterable, name=name, len_=len_)

##__________________________________________________________________||
class Atpbar(object):
    def __init__(self, iterable, name, len_):
        self.iterable = iterable
        self.name = name
        self.len_ = len_
        self.id_ = uuid.uuid4()

    def __iter__(self):
        self._report_progress(-1)
        for i, e in enumerate(self. iterable):
            yield e
            self._report_progress(i)

    def _report_progress(self, i):
        try:
            report = ProgressReport(
                name=self.name, done=(i+1),
                total=self.len_, taskid=self.id_)
            alphatwirl.progressbar.report_progress(report)
        except:
            pass

##__________________________________________________________________||



from .TableConfigCompleter import TableConfigCompleter
from .TableFileNameComposer import TableFileNameComposer
from .build_counter_collector_pair import build_counter_collector_pair
from .build_progressMonitor_communicationChannel import build_progressMonitor_communicationChannel



#!/usr/bin/env python
# Tai Sakuma <sakuma@cern.ch>
from __future__ import print_function
import os
import argparse

from HTCondorJobSubmitter import HTCondorJobSubmitter

##__________________________________________________________________||
import logging
logging.basicConfig(level = logging.DEBUG)

##__________________________________________________________________||
parser = argparse.ArgumentParser()
parser.add_argument('paths', nargs = argparse.REMAINDER, help = 'paths to task packages')
args = parser.parse_args()

##__________________________________________________________________||
def main():

    jobrunner = HTCondorJobSubmitter()

    for package_path_org in args.paths:
        package_path_abs = os.path.abspath(package_path_org)
        taskdir = os.path.dirname(package_path_abs)
        package_path_rel = os.path.basename(package_path_abs)

        jobrunner.run(taskdir, package_path_rel)

    try:
        jobrunner.wait()
    except KeyboardInterrupt:
        print('received KeyboardInterrupt')
        jobrunner.terminate()

##__________________________________________________________________||
if __name__ == '__main__':
    main()



# Tai Sakuma <tai.sakuma@gmail.com>

import os, errno
import logging

##__________________________________________________________________||
def mkdir_p(path):

    if not path:
        return

    logger = logging.getLogger(__name__)

    # http://stackoverflow.com/questions/600268/mkdir-p-functionality-in-python
    try:
        os.makedirs(path)
        logger.info('created a directory, {}'.format(path))

    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            loglevel = logging.getLevelName('DEBUG') - 1 # one level below DEBUG
            logger.log(loglevel, 'tried to create a directory, {}. already existed'.format(path))
            pass
        else:
            raise

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import os
import logging
import subprocess
import collections

##__________________________________________________________________||
class SubprocessRunner(object):
    def __init__(self, pipe=False):
        self.running_procs = collections.deque()
        self.pipe = pipe
        self.finished_pids = [ ]

    def __repr__(self):
        name_value_pairs = (
            ('pipe', self.pipe),
        )
        return '{}({})'.format(
            self.__class__.__name__,
            ', '.join(['{}={!r}'.format(n, v) for n, v in name_value_pairs]),
        )

    def run(self, workingArea, package_index):

        taskdir = workingArea.path

        package_path = workingArea.package_path(package_index)

        # run_script = os.path.join(taskdir, 'run.py') # This doesn't work.
                                                       # It contradicts with the document https://docs.python.org/2/library/subprocess.html
                                                       # The program's path needs to be relative to cwd
        run_script = os.path.join('.', 'run.py') # This works

        args = [run_script, package_path]
        proc = subprocess.Popen(
            args,
            stdout=subprocess.PIPE if self.pipe else None,
            stderr=subprocess.PIPE if self.pipe else None,
            cwd=taskdir
        )
        self.running_procs.append(proc)
        return proc.pid # as runid

    def run_multiple(self, workingArea, package_indices):
        pids = [ ]
        for pkgidx in package_indices:
            pids.append(self.run(workingArea, pkgidx))
        return pids

    def poll(self):
        """check if the jobs are running and return a list of pids for
        finished jobs

        """
        finished_procs = [p for p in self.running_procs if p.poll() is not None]
        self.running_procs = collections.deque([p for p in self.running_procs if p not in finished_procs])

        for proc in finished_procs:
            stdout, stderr = proc.communicate()
            ## proc.communicate() returns (stdout, stderr) when
            ## self.pipe = True. Otherwise they are (None, None)

        finished_pids = [p.pid for p in finished_procs]
        self.finished_pids.extend(finished_pids)

        logger = logging.getLogger(__name__)
        messages = 'Running: {}, Finished: {}'.format(len(self.running_procs), len(self.finished_pids))
        logger.info(messages)

        return finished_pids # as runids

    def wait(self):
        """wait until all jobs finish and return a list of pids
        """
        finished_pids = [ ]
        while self.running_procs:
            finished_pids.extend(self.poll())
        return finished_pids # as runids

    def failed_runids(self, runids):
        pass

    def terminate(self):
        while self.running_procs:
            proc = self.running_procs.popleft()
            proc.terminate()

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

from alphatwirl.selection.modules import All, Any, Not
from alphatwirl.selection.modules import AllwCount, AnywCount, NotwCount

##__________________________________________________________________||
allany_classes = [All, AllwCount, Any, AnywCount]
allany_classe_ids = [c.__name__ for c in allany_classes]

@pytest.mark.parametrize('Class', allany_classes, ids=allany_classe_ids)
def test_allany_begin(Class):
    obj = Class()
    sel1 = mock.Mock()
    sel2 = mock.Mock()
    del sel2.begin
    obj.add(sel1)
    obj.add(sel2)
    event = mock.Mock()
    obj.begin(event)
    assert [mock.call(event)] == sel1.begin.call_args_list

@pytest.mark.parametrize('Class', allany_classes, ids=allany_classe_ids)
def test_allany_end(Class):
    obj = Class()
    sel1 = mock.Mock()
    sel2 = mock.Mock()
    del sel2.end
    obj.add(sel1)
    obj.add(sel2)
    obj.end()
    assert [mock.call()] == sel1.end.call_args_list

##__________________________________________________________________||
not_classes = [Not, NotwCount]
not_classe_ids = [c.__name__ for c in not_classes]

@pytest.mark.parametrize('Class', not_classes, ids=not_classe_ids)
def test_not_begin(Class):
    selection = mock.Mock()
    obj = Class(selection)
    event = mock.Mock()
    obj.begin(event)
    assert [mock.call(event)] == selection.begin.call_args_list

@pytest.mark.parametrize('Class', not_classes, ids=not_classe_ids)
def test_not_begin_absent(Class):
    selection = mock.Mock()
    del selection.begin
    obj = Class(selection)
    event = mock.Mock()
    obj.begin(event)

@pytest.mark.parametrize('Class', not_classes, ids=not_classe_ids)
def test_not_end(Class):
    selection = mock.Mock()
    obj = Class(selection)
    obj.end()
    assert [mock.call()] == selection.end.call_args_list

@pytest.mark.parametrize('Class', not_classes, ids=not_classe_ids)
def test_not_end_absent(Class):
    selection = mock.Mock()
    del selection.end
    obj = Class(selection)
    obj.end()

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import itertools

##__________________________________________________________________||
class ToTupleList(object):
    def __init__(self, summaryColumnNames
                 ):

        self.summaryColumnNames = summaryColumnNames

    def __repr__(self):

        name_value_pairs = (
            ('summaryColumnNames', self.summaryColumnNames),
        )
        return '{}({})'.format(
            self.__class__.__name__,
            ', '.join(['{} = {!r}'.format(n, v) for n, v in name_value_pairs]),
        )

    def combine(self, dataset_readers_list):


        if len(dataset_readers_list) == 0: return None

        # e.g.,
        # dataset_readers_list = [
        #     ('QCD',    (reader1, reader2)),
        #     ('TTJets', (reader3, )),
        #     ('WJets',  (reader4, )),
        #     ('ZJets',  ( )),
        # ]

        readers_list = itertools.chain(*(r for _, r in dataset_readers_list))
        # e.g.,
        # readers_list = (reader1, reader2, reader3, reader4)

        summarizers_list = (r.results() for r in readers_list)
        # e.g.,
        # summarizers_list = (summarizer1, summarizer2, summarizer3, summarizer4)

        summarizer = sum(summarizers_list)

        ret = summarizer.to_tuple_list()
        # e.g.,
        # ret = [
        #         (200, 2, 120, 240),
        #         (300, 2, 490, 980),
        #         (300, 3, 210, 420)
        #         (300, 2, 20, 40),
        #         (300, 3, 15, 30)
        # ]

        ret.insert(0, self.summaryColumnNames)
        # e.g.,
        # [
        #     ('htbin', 'njetbin', 'n', 'nvar'),
        #     (    200,         2, 120,    240),
        #     (    300,         2, 490,    980),
        #     (    300,         3, 210,    420),
        #     (    300,         2,  20,     40),
        #     (    300,         3,  15,     30)
        # ]

        return ret

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import logging
import pytest

from alphatwirl.configure import build_progressMonitor_communicationChannel

##__________________________________________________________________||
def test_build_depricated(caplog):
    with pytest.raises(RuntimeError):
        with caplog.at_level(logging.ERROR):
            ret = build_progressMonitor_communicationChannel(quiet=True, processes=4)

    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'ERROR'
    assert 'alphatwirl.configure' in caplog.records[0].name
    assert 'removed' in caplog.records[0].msg

##__________________________________________________________________||



import itertools
import copy

from .Count import Count

##__________________________________________________________________||
class WithCountBase(object):
    """The base class of the classes AllwCount and AnywCount

    """

    def __init__(self, name, selections, collector):
        self.name = name
        if selections is None:
            selections = [ ]
        self.selections = list(selections)
        self.count = Count(selections=self.selections)
        self.collector = collector

        self._repr_pairs = [
            ('name', self.name),
            ('selections', self.selections),
            ('count', self.count),
        ]

    def __repr__(self):
        return '{}({})'.format(
            self.__class__.__name__,
            ', '.join(['{}={!r}'.format(n, v) for n, v in self._repr_pairs]),
        )

    def __str__(self):
        nwidth = max(len(n) for n, _ in self._repr_pairs)
        nwidth += 4
        nwidth_nested = 12
        lines = [
            '{}:'.format(self.__class__.__name__),
            '{:>{}}: {!r}'.format('name', nwidth, self.name),
            '{:>{}}: {!r}'.format('count', nwidth, self.count),
            '{:>{}}:'.format('selections', nwidth),
        ]
        nested_lines = list(itertools.chain(*[str(s).split('\n') for s in self.selections]))
        lines.extend(
            ['{}{}'.format(' '*nwidth_nested, str(l)) for l in nested_lines]
        )
        return '\n'.join(lines)

    def add(self, selection):
        self.selections.append(selection)
        self.count.add(selection)

    def begin(self, event):
        for s in self.selections:
            if hasattr(s, 'begin'): s.begin(event)

    def event(self, event):
        return self(event)

    def end(self):
        for s in self.selections:
            if hasattr(s, 'end'): s.end()

    def merge(self, other):
        self.count += other.count
        for s, o in zip(self.selections, other.selections):
            if not hasattr(s, 'merge'):
                continue
            s.merge(o)

    def results(self, increment=False):

        ret = self.count.copy()

        # reversed enumerate
        for i, s in zip(reversed(range(len(self.selections))), reversed(self.selections)):
            if hasattr(s, 'results'):
                ret.insert(i, s.results(increment=True))

        if increment:
            ret.increment_depth(by=1)

        return ret

    def collect(self):
        if self.collector is not None:
            return self.collector(self)

##__________________________________________________________________||
class AllwCount(WithCountBase):
    """select events that meet all conditions

    """

    def __init__(self, name='All', selections=None, collector=None):
        if name is None:
            name = 'All'
        super(AllwCount, self).__init__(name, selections, collector)

    def __call__(self, event):
        ret = True
        pass_ = [ ]
        for s in self.selections:
            pass_.append(s(event))
            if not pass_[-1]:
                ret = False
                break
        self.count.count(pass_)
        return ret

##__________________________________________________________________||
class AnywCount(WithCountBase):
    """select events that meet any of the conditions

    """

    def __init__(self, name='Any', selections=None, collector=None):
        if name is None:
            name = 'Any'
        super(AnywCount, self).__init__(name, selections, collector)

    def __call__(self, event):
        ret = False
        pass_ = [ ]
        for s in self.selections:
            pass_.append(s(event))
            if pass_[-1]:
                ret = True
                break
        self.count.count(pass_)
        return ret

##__________________________________________________________________||
class NotwCount(object):
    """select events that do NOT pass the selection

    """

    def __init__(self, selection, name='Not', collector=None):
        if name is None:
            name = 'Not'
        self.name = name
        self.selection = selection
        self.count = Count(selections=[self.selection])
        self.collector = collector

    def __repr__(self):
        return '{}(name={!r}, selection={!r}), count={!r}'.format(
            self.__class__.__name__,
            self.name,
            self.selection,
            self.count
        )

    def begin(self, event):
        if hasattr(self.selection, 'begin'): self.selection.begin(event)

    def __call__(self, event):
        pass_ = self.selection(event)
        self.count.count([pass_])
        return not pass_

    def event(self, event):
        return self(event)

    def end(self):
        if hasattr(self.selection, 'end'): self.selection.end()

    def merge(self, other):
        self.count += other.count
        if not hasattr(self.selection, 'merge'):
            return
        self.selection.merge(other.selection)

    def results(self, increment=False):
        ret = self.count.copy()
        if hasattr(self.selection, 'results'):
            ret.insert(0, self.selection.results(increment=True))
        if increment:
            ret.increment_depth(by=1)
        return ret

    def collect(self):
        if self.collector is not None:
            return self.collector(self)

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import logging

from ..progressbar import ProgressReport

from alphatwirl.misc.deprecation import _deprecated_class_method_option

import alphatwirl


##__________________________________________________________________||
class CollectorComposite(object):

    """A composite of collectors.

    This class is a composite in the composite pattern.

    Examples of collectors are instances of `Collector`,
    `NullCollector`, and this class.

    """

    @_deprecated_class_method_option('progressReporter')
    def __init__(self, progressReporter=None):

        self.components = [ ]

    def __repr__(self):
        name_value_pairs = (
            ('components',       self.components),
        )
        return '{}({})'.format(
            self.__class__.__name__,
            ', '.join(['{}={!r}'.format(n, v) for n, v in name_value_pairs]),
        )

    def add(self, collector):
        """add a collector


        Args:
            collector: the collector to be added

        """
        self.components.append(collector)

    def collect(self, dataset_readers_list):
        """collect results


        Returns:
            a list of results

        """

        ret = [ ]
        for i, collector in enumerate(self.components):
            report = ProgressReport(name='collecting results', done=(i + 1), total=len(self.components))
            alphatwirl.progressbar.report_progress(report)
            ret.append(collector.collect([(dataset, tuple(r.readers[i] for r in readerComposites))
                                          for dataset, readerComposites in dataset_readers_list]))
        return ret

        ## in one line without the progress bar
        #return [collector.collect([(dataset, tuple(r.readers[i] for r in readerComposites))
        #                           for dataset, readerComposites in dataset_readers_list])
        #        for i, collector in enumerate(self.components)]

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import sys
import numpy as np
import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

from alphatwirl.summary import Summarizer
from alphatwirl.summary import Sum

##__________________________________________________________________||
@pytest.fixture()
def obj():
    return Summarizer(Summary=Sum)

def test_repr(obj):
    repr(obj)

def test_add(obj):

    obj.add('A', (12, ) )
    expected  = {'A': Sum(contents=np.array((12, )))}
    assert expected == obj.results()

    obj.add('A', (23, ))
    expected  = {'A': Sum(contents=np.array((35, )))}
    assert expected == obj.results()

    obj.add('A', (10, ), weight=2)
    expected  = {'A': Sum(contents=np.array((55, )))}
    assert expected == obj.results()

    obj.add('B', (20, ), weight=3.2)
    expected  = {
        'A': Sum(contents=np.array((55, ))),
        'B': Sum(contents=np.array((64.0, ))),
    }
    assert expected == obj.results()

def test_add_key(obj):

    obj.add_key('A')
    expected  = {'A': Sum(contents=np.array((0, )))}
    assert expected == obj.results()

    obj.add_key('B')
    obj.add_key('C')
    expected  = {
        'A': Sum(contents=np.array((0, ))),
        'B': Sum(contents=np.array((0, ))),
        'C': Sum(contents=np.array((0, ))),
    }
    assert expected == obj.results()

def test_key(obj):
    obj.add_key('A')
    assert ['A'] == list(obj.keys())

def test_to_tuple_list(obj):
    obj.add(('A', ), (12, ))
    obj.add(('B', ), (20, ))
    assert [('A', 12), ('B', 20)] == obj.to_tuple_list()

@pytest.mark.skipif(sys.version_info >= (3, 0), reason="requires python 2")
def test_to_tuple_list_key_not_tuple(obj):
    obj.add('A', (12, )) # the keys are not a tuple
    obj.add(2, (20, ))   #
    assert [(2, 20), ('A', 12)] == obj.to_tuple_list()

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import pytest

from alphatwirl.selection.factories.factory import FactoryDispatcher
from alphatwirl.selection.modules.LambdaStr import LambdaStr
from alphatwirl.selection.modules import All, Any, Not

##__________________________________________________________________||
alias_dict = {
    'alias1': 'ev : ev.var1[0] >= 10',
    'alias2': ('ev : ev.var2[0] >= 20', dict(name='name2')),
    'alias3': 'alias1',
    'alias4': 'alias3',
    'alias5': 'ev : ev.var4[0] == {n}',
    'alias6': ('ev : {low} <= ev.var5[0] < {high}', dict(low=11, high=20))
}

##__________________________________________________________________||
params = [
    pytest.param(
        dict(Any=(
            'ev : ev.x[0] == 0',
            dict(All=(
                'ev : ev.x[0] >= 1',
                'ev : ev.y[0] >= 100',
                'alias4',
                )),
            dict(Not=dict(
                Any=(
                    'ev : ev.z[0] == 0',
                    'ev : ev.w[0] >= 300',
                    'alias5',
                    'alias6',
                ),
            )),
        )),
        dict(
            AllClass=All, AnyClass=Any, NotClass=Not,
            LambdaStrClass=LambdaStr, aliasDict=alias_dict,
            n=5242,
        ),
        Any(selections=[
            LambdaStr('ev : ev.x[0] == 0'),
            All(selections=[
                LambdaStr('ev : ev.x[0] >= 1'),
                LambdaStr('ev : ev.y[0] >= 100'),
                LambdaStr(name='alias4', lambda_str='ev : ev.var1[0] >= 10'),
            ]),
            Not(selection=Any(selections=[
                LambdaStr('ev : ev.z[0] == 0'),
                LambdaStr('ev : ev.w[0] >= 300'),
                LambdaStr(name='alias5', lambda_str='ev : ev.var4[0] == 5242'),
                LambdaStr(name='alias6', lambda_str='ev : 11 <= ev.var5[0] < 20'),
            ]),
            ),
        ]),
        id='example'
    ),
]

@pytest.mark.parametrize('path_cfg, kargs, expected', params)
def test_dispatcher(path_cfg, kargs, expected):
    obj = FactoryDispatcher(path_cfg=path_cfg, **kargs)
    assert repr(expected) == repr(obj)
    assert str(expected) == str(obj)

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import sys

import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

from alphatwirl.progressbar import BProgressMonitor, ProgressReporter, ProgressReport

import alphatwirl

##__________________________________________________________________||
@pytest.fixture()
def presentation():
    ret = mock.MagicMock()
    return ret

@pytest.fixture()
def monitor(presentation):
    return BProgressMonitor(presentation)


##__________________________________________________________________||
def test_repr(monitor):
    repr(monitor)

def test_begin_end(monitor, presentation):
    presentation.nreports.return_value = 0
    monitor.begin()
    assert isinstance(alphatwirl.progressbar._progress_reporter, ProgressReporter)
    monitor.end()
    assert alphatwirl.progressbar._progress_reporter is None

def test_createReporter(monitor):
    reporter = monitor.createReporter()
    assert isinstance(reporter, ProgressReporter)

def test_send_report(monitor, presentation):
    presentation.nreports.return_value = 10
    monitor.begin()
    reporter = monitor.createReporter()
    report = ProgressReport('task1', 0, 3)
    reporter.report(report)
    monitor.end()

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import pytest

from alphatwirl.roottree import Events

##__________________________________________________________________||
class MockFile(object):
    pass

class MockTree(object):
    def __init__(self, entries=100):
        self.entries = entries
        self.ievent = -1
        self.branchstatus = [ ]
        self.branch1 = 1111
        self.directory = MockFile()
    def GetDirectory(self):
        return self.directory
    def GetEntries(self):
        return self.entries
    def GetEntry(self, ientry):
        if ientry < self.entries:
            nbytes = 10
            self.ievent = ientry
        else:
            nbytes = 0
            self.ievent = -1
        return nbytes

    def SetBranchStatus(self, bname, status):
        self.branchstatus.append((bname, status))

def test_mocktree():

    tree = MockTree(entries=3)
    assert isinstance(tree.GetDirectory(), MockFile)
    assert 3 == tree.GetEntries()

    assert -1 == tree.ievent

    nbytes = 10
    assert nbytes == tree.GetEntry(0)
    assert 0 == tree.ievent
    assert nbytes == tree.GetEntry(1)
    assert 1 == tree.ievent
    assert nbytes == tree.GetEntry(2)
    assert 2 == tree.ievent
    assert 0 == tree.GetEntry(3)
    assert -1 == tree.ievent

##__________________________________________________________________||
def test_iter_getattr():
    tree = MockTree(entries=5)
    events = Events(tree)
    it = iter(events)
    event = next(it)
    assert 1111 == event.branch1
    tree.branch1 = 2222
    assert 2222 == event.branch1

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
from __future__ import print_function
import logging

import pytest

try:
   import cPickle as pickle
except:
   import pickle

from alphatwirl.misc.deprecation import _deprecated_func_option
from alphatwirl.misc.deprecation import _deprecated_class_method_option

##__________________________________________________________________||
@_deprecated_func_option('B')
def func_01(A=123, B=456, C=None):
    pass

def test_func_option_not_used(caplog):

    with caplog.at_level(logging.WARNING, logger='alphatwirl'):
        func_01(A=490)

    assert len(caplog.records) == 0

def test_func_option_used(caplog):

    with caplog.at_level(logging.WARNING, logger='alphatwirl'):
        func_01(B=789)

    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'WARNING'
    assert 'test_deprecation' in caplog.records[0].name
    assert 'func_01(): the option "B" is deprecated.' in caplog.records[0].msg

def test_func_pickle():
    pickle.dumps(func_01)

##__________________________________________________________________||
@_deprecated_func_option('C')
def func_02(A=123, B=456, C=None):
    pass

def test_func_option_used_none(caplog):

    with caplog.at_level(logging.WARNING, logger='alphatwirl'):
        func_02(C=None)

    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'WARNING'
    assert 'test_deprecation' in caplog.records[0].name
    assert 'func_02(): the option "C" is deprecated.' in caplog.records[0].msg

##__________________________________________________________________||
@_deprecated_func_option('B')
@_deprecated_func_option('C')
def func_03(A=123, B=456, C=None):
    pass

def test_func_option_used_double_B(caplog):

    with caplog.at_level(logging.WARNING, logger='alphatwirl'):
        func_03(B=4)

    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'WARNING'
    assert 'test_deprecation' in caplog.records[0].name
    assert 'func_03(): the option "B" is deprecated.' in caplog.records[0].msg

def test_func_option_used_double_C(caplog):

    with caplog.at_level(logging.WARNING, logger='alphatwirl'):
        func_03(C=None)

    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'WARNING'
    assert 'test_deprecation' in caplog.records[0].name
    assert 'func_03(): the option "C" is deprecated.' in caplog.records[0].msg

def test_func_option_used_double_BC(caplog):

    with caplog.at_level(logging.WARNING, logger='alphatwirl'):
        func_03(B=1, C=3)

    assert len(caplog.records) == 2
    assert caplog.records[0].levelname == 'WARNING'
    assert caplog.records[1].levelname == 'WARNING'
    assert 'test_deprecation' in caplog.records[0].name
    assert 'test_deprecation' in caplog.records[1].name
    assert 'func_03(): the option "B" is deprecated.' in caplog.records[0].msg
    assert 'func_03(): the option "C" is deprecated.' in caplog.records[1].msg

##__________________________________________________________________||
class Class(object):
    @_deprecated_class_method_option('B')
    def __init__(A=123, B=456, C=None):
        pass

def test_class_init_option_used(caplog):

    with caplog.at_level(logging.WARNING, logger='alphatwirl'):
        Class(B=1)

    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'WARNING'
    assert 'test_deprecation' in caplog.records[0].name
    assert 'tests.unit.misc.test_deprecation_option.Class.__init__(): the option "B" is deprecated.' in caplog.records[0].msg

def test_class_pickle():
    c = Class()
    pickle.dumps(c)
##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import numpy as np
import copy

import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

from alphatwirl.summary import Count

##__________________________________________________________________||
@pytest.mark.parametrize('kwargs, expected_contents', (
    (dict(), [np.array([0, 0])]),
    (dict(val=()), [np.array([1, 1])]),
    (dict(val=(), weight=10), [np.array([10, 100])]),
    (dict(contents=[np.array([1, 3])]), [np.array([1, 3])]),
))
def test_init(kwargs, expected_contents):
    obj = Count(**kwargs)
    np.testing.assert_equal(expected_contents, obj.contents)

##__________________________________________________________________||
def test_repr():
    obj = Count()
    repr(obj)

##__________________________________________________________________||
def test_add():
    obj1 = Count(contents=[np.array((10, 20))])
    obj2 = Count(contents=[np.array((30, 40))])
    obj3 =  obj1 + obj2
    np.testing.assert_equal([np.array([40, 60])], obj3.contents)
    assert obj1 is not obj3
    assert obj1.contents is not obj3.contents
    assert obj2 is not obj3
    assert obj2.contents is not obj3.contents

def test_radd():
    obj1 = Count(contents=[np.array((10, 20))])
    assert obj1 is not sum([obj1]) # will call 0 + obj1
    assert obj1 == sum([obj1])

def test_radd_raise():
    obj1 = Count(contents=[np.array((10, 20))])
    with pytest.raises(TypeError):
        1 + obj1

def test_copy():
    obj1 = Count(contents=[np.array((10, 20))])
    copy1 = copy.copy(obj1)
    assert obj1 == copy1
    assert obj1 is not copy1
    assert obj1.contents is not copy1.contents
    assert obj1.contents[0] is not copy1.contents[0]

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import logging
import pytest

import alphatwirl
from alphatwirl.progressbar import atpbar

try:
    import unittest.mock as mock
except ImportError:
    import mock

##__________________________________________________________________||
@pytest.fixture()
def mock_report_progress(monkeypatch):
    ret = mock.Mock()
    monkeypatch.setattr(alphatwirl.progressbar, 'report_progress', ret)
    return ret

##__________________________________________________________________||
class Iter(object):
    def __init__(self, content):
        self.content = content

    def __iter__(self):
        for e in self.content:
            yield e

##__________________________________________________________________||
content = [mock.sentinel.item1, mock.sentinel.item2, mock.sentinel.item3]

##__________________________________________________________________||
def test_atpbar_no_len(mock_report_progress, caplog):
    iterable = Iter(content)

    ##
    returned = [ ]
    with caplog.at_level(logging.WARNING):
        for e in atpbar(iterable):
            returned.append(e)

    ##
    assert content == returned

    ##
    assert not mock_report_progress.call_args_list

    ##
    assert 2 == len(caplog.records)
    assert 'WARNING' == caplog.records[0].levelname
    assert 'length is unknown' in caplog.records[0].msg

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import copy

from .EventLoop import EventLoop
from alphatwirl.misc.deprecation import _deprecated

##__________________________________________________________________||
@_deprecated(msg='use EventDatasetReader instead.')
class EventsInDatasetReader(object):
    """This class manages objects involved in reading events in data sets.

    On receiving a data set, this class calls the function
    split_into_build_events(), which splits the data set into chunks,
    creates the function build_events() for each chunk, and returns a
    list of the functions. Then, for each build_events(), This class
    creates a copy of the reader, creates an event loop, and send it
    to the event loop runner.

    At the end, this class receives results from the event loop runner
    and have the collector collect them.

    """
    def __init__(self, eventLoopRunner, reader, collector,
                 split_into_build_events):

        self.eventLoopRunner = eventLoopRunner
        self.reader = reader
        self.collector = collector
        self.split_into_build_events = split_into_build_events

        self.EventLoop = EventLoop

        self.dataset_nreaders = [ ]

    def __repr__(self):
        name_value_pairs = (
            ('eventLoopRunner',         self.eventLoopRunner),
            ('reader',                  self.reader),
            ('collector',               self.collector),
            ('split_into_build_events', self.split_into_build_events),
        )
        return '{}({})'.format(
            self.__class__.__name__,
            ', '.join(['{}={!r}'.format(n, v) for n, v in name_value_pairs]),
        )

    def begin(self):
        self.eventLoopRunner.begin()
        self.dataset_nreaders = [ ]

    def read(self, dataset):
        build_events_list = self.split_into_build_events(dataset)
        self.dataset_nreaders.append((dataset, len(build_events_list)))
        eventLoops = [ ]
        for build_events in build_events_list:
            reader = copy.deepcopy(self.reader)
            eventLoop = self.EventLoop(build_events, reader, dataset.name)
            eventLoops.append(eventLoop)
        self.eventLoopRunner.run_multiple(eventLoops)

    def end(self):
        returned_readers = self.eventLoopRunner.end()

        nreaders_total = sum((n for _, n in self.dataset_nreaders))

        if nreaders_total != len(returned_readers):
            import logging
            logger = logging.getLogger(__name__)
            logger.warning(
                'the same number of the readers were not returned: {} readers sent, {} readers returned. cannot collect results'.format(
                    nreaders_total,
                    len(returned_readers)
                ))
            return None

        dataset_readers_list = [ ]
        i = 0
        for dataset, nreaders in self.dataset_nreaders:
            dataset_readers_list.append((dataset.name, tuple(returned_readers[i:(i + nreaders)])))
            i += nreaders
        return self.collector.collect(dataset_readers_list)

##__________________________________________________________________||



from .CommunicationChannel import CommunicationChannel
from .CommunicationChannel0 import CommunicationChannel0
from .TaskPackage import TaskPackage
from .TaskPackageDropbox import TaskPackageDropbox
from .MultiprocessingDropbox import MultiprocessingDropbox
from .SubprocessRunner import SubprocessRunner
from .WorkingArea import WorkingArea
from .HTCondorJobSubmitter import HTCondorJobSubmitter



import unittest
import copy

from alphatwirl.summary import Count, KeyValueComposer, NextKeyComposer, Reader
from alphatwirl.binning import Echo
from alphatwirl.loop import ReaderComposite
from alphatwirl.loop import Collector, CollectorComposite, CollectorDelegate

##__________________________________________________________________||
class MockResultsCombinationMethod(object):
    def combine(self, pairs) :pass

##__________________________________________________________________||
class TesEventsInDatasetReader_build_01(unittest.TestCase):

    def test_two(self):
        """
        1:composite
            |- 3:composite
            |  |- 4:counter
            |  |- 5:counter
            |
            |- 7:counter
            |- 8:counter
        """

        keyComposer4 = KeyValueComposer(('var4', ), (Echo(), ))
        counts4 = Count()
        reader4 = Reader(keyComposer4, counts4)
        collector4 = Collector(MockResultsCombinationMethod())

        keyComposer5 = KeyValueComposer(('var5', ), (Echo(), ))
        counts5 = Count()
        reader5 = Reader(keyComposer5, counts5)
        collector5 = Collector(MockResultsCombinationMethod())

        keyComposer7 = KeyValueComposer(('var7', ), (Echo(), ))
        counts7 = Count()
        reader7 = Reader(keyComposer7, counts7)
        collector7 = Collector(MockResultsCombinationMethod())

        keyComposer8 = KeyValueComposer(('var8', ), (Echo(), ))
        counts8 = Count()
        reader8 = Reader(keyComposer8, counts8)
        collector8 = Collector(MockResultsCombinationMethod())

        reader3 = ReaderComposite()
        reader3.add(reader4)
        reader3.add(reader5)

        collector3 = CollectorComposite()
        collector3.add(collector4)
        collector3.add(collector5)

        reader1 = ReaderComposite()
        reader1.add(reader3)
        reader1.add(reader7)
        reader1.add(reader8)

        collector1 = CollectorComposite()
        collector1.add(collector3)
        collector1.add(collector7)
        collector1.add(collector8)


        reader1_ds1 = copy.deepcopy(reader1)
        reader1_ds2 = copy.deepcopy(reader1)

        reader3_ds1 = reader1_ds1.readers[0]
        reader4_ds1 = reader3_ds1.readers[0]
        reader5_ds1 = reader3_ds1.readers[1]
        reader7_ds1 = reader1_ds1.readers[1]
        reader8_ds1 = reader1_ds1.readers[2]

        self.assertIsInstance(reader1_ds1, ReaderComposite)
        self.assertIsInstance(reader3_ds1, ReaderComposite)
        self.assertIsInstance(reader4_ds1, Reader)
        self.assertIsInstance(reader5_ds1, Reader)
        self.assertIsInstance(reader7_ds1, Reader)
        self.assertIsInstance(reader8_ds1, Reader)

        self.assertIsNot(reader1, reader1_ds1)
        self.assertIsNot(reader3, reader3_ds1)
        self.assertIsNot(reader4, reader4_ds1)
        self.assertIsNot(reader5, reader5_ds1)
        self.assertIsNot(reader7, reader7_ds1)
        self.assertIsNot(reader8, reader8_ds1)

        reader3_ds2 = reader1_ds2.readers[0]
        reader4_ds2 = reader3_ds2.readers[0]
        reader5_ds2 = reader3_ds2.readers[1]
        reader7_ds2 = reader1_ds2.readers[1]
        reader8_ds2 = reader1_ds2.readers[2]

        self.assertIsInstance(reader1_ds2, ReaderComposite)
        self.assertIsInstance(reader3_ds2, ReaderComposite)
        self.assertIsInstance(reader4_ds2, Reader)
        self.assertIsInstance(reader5_ds2, Reader)
        self.assertIsInstance(reader7_ds2, Reader)
        self.assertIsInstance(reader8_ds2, Reader)

        self.assertIsNot(reader1, reader1_ds2)
        self.assertIsNot(reader3, reader3_ds2)
        self.assertIsNot(reader4, reader4_ds2)
        self.assertIsNot(reader5, reader5_ds2)
        self.assertIsNot(reader7, reader7_ds2)
        self.assertIsNot(reader8, reader8_ds2)

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
from __future__ import print_function
import logging

import pytest

try:
    import cPickle as pickle
except:
    import pickle

from alphatwirl.misc.deprecation import _renamed_class_method_option
from alphatwirl.misc.deprecation import _renamed_func_option

##__________________________________________________________________||
@_renamed_func_option(old='B', new='C')
def func_01(A=123, C=None):
    return dict(A=A, C=C)

def test_func_option_used(caplog):

    with caplog.at_level(logging.WARNING):
        r = func_01(B=455)

    assert r['C'] == 455

    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'WARNING'
    assert 'test_deprecation' in caplog.records[0].name
    assert 'renamed' in caplog.records[0].msg

def test_func_option_notused(caplog):

    with caplog.at_level(logging.WARNING):
        r = func_01(C=455)

    assert r['C'] == 455

    assert len(caplog.records) == 0

def test_class_pickle():
    c = Class()
    pickle.dumps(c)

##__________________________________________________________________||

##__________________________________________________________________||
class Class(object):
    @_renamed_class_method_option(old='B', new='C')
    def __init__(self, A=123, C=None):
        self.A = A
        self.C = C

def test_class_init_option_used(caplog):

    with caplog.at_level(logging.WARNING):
        c = Class(B=455)

    assert c.C == 455

    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'WARNING'
    assert 'test_deprecation' in caplog.records[0].name
    assert 'renamed' in caplog.records[0].msg

def test_class_init_option_notused(caplog):

    with caplog.at_level(logging.WARNING):
        c = Class(C=455)

    assert c.C == 455

    assert len(caplog.records) == 0

def test_class_pickle():
    c = Class()
    pickle.dumps(c)

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
from __future__ import print_function
import inspect
import functools
import logging

##__________________________________________________________________||
def _deprecated(msg=''):

    def _decorate_class(c, msg):
        module_name = c.__module__
        logger = logging.getLogger(module_name)
        class_name = c.__name__
        name = '{}.{}'.format(module_name, class_name)
        text = '{} is deprecated.'.format(name)
        if msg:
            text += ' ' + msg

        original_init = c.__init__

        def init(*args, **kwargs):
            logger.warning(text)
            return original_init(*args, **kwargs)

        c.__init__ = init
        return c

    def _decorate_func(f, msg):
        module_name = f.__module__
        logger = logging.getLogger(module_name)
        name = f.__name__
        name = '{}.{}'.format(module_name, name)
        text = '{}() is deprecated.'.format(name)
        if msg:
            text += ' ' + msg

        @functools.wraps(f)
        def g(*args, **kwargs):
            logger.warning(text)
            return f(*args, **kwargs)
        return g

    def _imp(f):
        if inspect.isclass(f):
            return _decorate_class(f, msg)
        return _decorate_func(f, msg)
    return _imp

##__________________________________________________________________||
def _deprecated_func_option(option, msg=''):
    def _imp(f):
        module_name = f.__module__
        logger = logging.getLogger(module_name)
        func_name = f.__name__

        @functools.wraps(f)
        def g(*args, **kwargs):
            if option in kwargs:
                name = '{}.{}'.format(module_name, func_name)
                text = '{}(): the option "{}" is deprecated.'.format(name, option)
                text += ' "{}={}" is given.'.format(option, kwargs[option])
                if msg:
                    text += ' ' + msg
                logger.warning(text)
            return f(*args, **kwargs)
        return g
    return _imp

##__________________________________________________________________||
def _deprecated_class_method_option(option, msg=''):
    def _imp(f):
        module_name = f.__module__
        logger = logging.getLogger(module_name)
        method_name = f.__name__
        def g(*args, **kwargs):
            if option in kwargs:
                name = '{}.{}.{}'.format(module_name, args[0].__class__.__name__, method_name)
                text = '{}(): the option "{}" is deprecated.'.format(name, option)
                text += ' "{}={}" is given.'.format(option, kwargs[option])
                if msg:
                    text += ' ' + msg
                logger.warning(text)
            return f(*args, **kwargs)
        return g
    return _imp

##__________________________________________________________________||
def _renamed_func_option(old, new, msg=''):
    def _imp(f):
        module_name = f.__module__
        logger = logging.getLogger(module_name)
        func_name = f.__name__
        def g(*args, **kwargs):
            if old in kwargs:
                name = '{}.{}'.format(module_name, func_name)
                text = '{}(): the option "{}" is renamed "{}".'.format(name, old, new)
                text += ' "{}={}" is given.'.format(old, kwargs[old])
                if msg:
                    text += ' ' + msg
                logger.warning(text)
                kwargs[new] = kwargs.pop(old)
            return f(*args, **kwargs)
        return g
    return _imp

##__________________________________________________________________||
def _renamed_class_method_option(old, new, msg=''):
    def _imp(f):
        module_name = f.__module__
        logger = logging.getLogger(module_name)
        method_name = f.__name__
        def g(*args, **kwargs):
            if old in kwargs:
                name = '{}.{}.{}'.format(module_name, args[0].__class__.__name__, method_name)
                text = '{}(): the option "{}" is renamed "{}".'.format(name, old, new)
                text += ' "{}={}" is given.'.format(old, kwargs[old])
                if msg:
                    text += ' ' + msg
                logger.warning(text)
                kwargs[new] = kwargs.pop(old)
            return f(*args, **kwargs)
        return g
    return _imp

##__________________________________________________________________||
from .removal import _removed

@_removed(msg='use _deprecated() instead')
def atdeprecated(*args, **kwargs):
    pass

@_removed(msg='use _deprecated_func_option() instead')
def atdeprecated_func_option(*args, **kwargs):
    pass

@_removed(msg='use _deprecated_class_method_option() instead')
def atdeprecated_class_method_option(*args, **kwargs):
    pass

@_removed(msg='use _renamed_func_option() instead')
def atrenamed_func_option(*args, **kwargs):
    pass

@_removed(msg='use _renamed_class_method_option() instead')
def atrenamed_class_method_option(*args, **kwargs):
    pass

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import logging

from .TaskPackage import TaskPackage

##__________________________________________________________________||
class CommunicationChannel(object):
    """A communication channel with workers in other processes.

    (This docstring is outdated.)

    You can send tasks to workers through this channel. The workers,
    running in other processes, execute the tasks in the background.
    You can receive the results of the tasks also through this
    channel.

    An instance of this class can be created with two optional
    arguments: ``nprocesses``, the number of workers to be created,
    and ``progressMonitor``::

        progressBar = ProgressBar()
        progressMonitor = BProgressMonitor(progressBar)
        channel = CommunicationChannel(nprocesses=10, progressMonitor=progressMonitor)

    Workers will be created when ``begin()`` is called::

        channel.begin()

    In ``begin()``, this class gives each worker a
    ``progressReporter``, which is created by the ``progressMonitor``.

    Now, you are ready to send a task. A task is a function or any
    object which is callable and picklable. A task can take any number
    of arguments. If a task takes a named argument
    ``progressReporter``, the worker will give the
    ``progressReporter`` to the task. A value that a task returns is
    the result of the task and must be picklable. For example, an
    instance of ``EventLoop`` can be a task. You can send a task with
    the method ``put``::

        channel.put(task1, 10, 20, A=30)

    Here, 10, 20, A=30 are the arguments to the task.

    This class sends the task to a worker. The worker which receives
    the task will first try to call the task with the
    ``progressReporter`` in addition to the arguments. If the task
    doesn't take the ``progressReporter``, it calls only with the
    arguments.

    You can send multiple tasks::

        channel.put(task2)
        channel.put(task3, 100, 200)
        channel.put(task4, A='abc')
        channel.put(task5)

    They will be executed by workers.

    You can receive the results of the tasks with the method
    ``receive()``::

        results = channel.receive()

    This method will wait until all tasks are finished. If a task
    gives a ``progressReport`` to the ``progressReporter``, the report
    will be used, for example, to update progress bars on the screen.

    When all tasks end, results will be returned. The return value
    ``results`` is a list of results of the tasks. They are sorted in
    the order in which the tasks are originally put.

    After receiving the results, you can put more tasks::

        channel.put(task6)
        channel.put(task7)

    And you can receive the results of them::

        results = channel.receive()

    If there are no more tasks to be done, you should call the method
    ``end``::

        channel.end()

    This will end all background processes.

    """

    def __init__(self, dropbox):
        self.dropbox = dropbox
        self.isopen = False

    def __repr__(self):
        return '{}(dropbox={!r}, isopen={!r})'.format(
            self.__class__.__name__,
            self.dropbox,
            self.isopen
        )

    def begin(self):
        if self.isopen: return
        self.dropbox.open()
        self.isopen = True

    def put(self, task, *args, **kwargs):
        if not self.isopen:
            logger = logging.getLogger(__name__)
            logger.warning('the drop box is not open')
            return
        package = TaskPackage(task=task, args=args, kwargs=kwargs)
        return self.dropbox.put(package)

    def put_multiple(self, task_args_kwargs_list):
        if not self.isopen:
            logger = logging.getLogger(__name__)
            logger.warning('the drop box is not open')
            return

        packages = [ ]
        for t in task_args_kwargs_list:
            try:
                task = t['task']
                args = t.get('args', ())
                kwargs = t.get('kwargs', {})
                package = TaskPackage(task=task, args=args, kwargs=kwargs)
            except TypeError:
                package = TaskPackage(task=t, args=(), kwargs={})
            packages.append(package)
        return self.dropbox.put_multiple(packages)

    def receive_finished(self):
        if not self.isopen:
            logger = logging.getLogger(__name__)
            logger.warning('the drop box is not open')
            return
        return self.dropbox.poll()

    def receive_one(self):
        if not self.isopen:
            logger = logging.getLogger(__name__)
            logger.warning('the drop box is not open')
            return
        return self.dropbox.receive_one()

    def receive_all(self):
        if not self.isopen:
            logger = logging.getLogger(__name__)
            logger.warning('the drop box is not open')
            return
        return self.dropbox.receive()

    def receive(self):
        pkgidx_result_pairs = self.receive_all()
        if pkgidx_result_pairs is None:
            return
        results = [r for _, r in pkgidx_result_pairs]
        return results

    def terminate(self):
        self.dropbox.terminate()

    def end(self):
        if not self.isopen: return
        self.dropbox.close()
        self.isopen = False

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
from .ProgressReporter import ProgressReporter

import alphatwirl

##__________________________________________________________________||
class Queue(object):

    def __init__(self, presentation):
        self.presentation = presentation

    def put(self, report):
        self.presentation.present(report)

##__________________________________________________________________||
class ProgressMonitor(object):
    def __init__(self, presentation):
        self.presentation = presentation
        self.queue = Queue(presentation=presentation)

    def begin(self):
        reporter = self.createReporter()
        alphatwirl.progressbar._progress_reporter = reporter

    def end(self):
        alphatwirl.progressbar._progress_reporter = None

    def createReporter(self):
        reporter = ProgressReporter(queue = self.queue)
        return reporter

##__________________________________________________________________||



#!/usr/bin/env python
import logging
import subprocess
import time

##__________________________________________________________________||
def try_executing_until_succeed(procargs, sleep=2):

    logger = logging.getLogger(__name__)

    while True:

        #
        command_display = compose_shortened_command_for_logging(procargs)
        logger.debug('execute: {!r}'.format(command_display))

        #
        proc = subprocess.Popen(
            procargs,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )
        stdout, stderr = proc.communicate()
        success = not (proc.returncode or stderr)

        #
        if success: break

        #
        if stderr: logger.warning(stderr.strip())
        logger.warning('the command failed: {!r}. will try again in {} seconds'.format(command_display, sleep))

        #
        time.sleep(sleep)

    if not stdout: return [ ]
    return stdout.rstrip().split(b'\n')

##__________________________________________________________________||
def compose_shortened_command_for_logging(procargs):
    ellipsis = '...(({} letters))...'
    nfirst = 50
    nlast = 50
    command_display = subprocess.list2cmdline(procargs)
    if len(command_display) > nfirst + len(ellipsis) + nlast:
        command_display = '{}...(({} letters))...{}'.format(
            command_display[:nfirst],
            len(command_display) - (nfirst + nlast),
            command_display[-nlast:]
        )
    return command_display

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
from ..misc import mkdir_p
from ..misc import list_to_aligned_text
import os

##__________________________________________________________________||
class WriteListToFile(object):
    def __init__(self, outPath):
        self._outPath = outPath

    def __repr__(self):
        return '{}(outPath={!r})'.format(
            self.__class__.__name__,
            self._outPath
        )

    def deliver(self, results):
        if results is None: return
        f = self._open(self._outPath)
        f.write(list_to_aligned_text(results).encode())
        self._close(f)

    def _open(self, path):
        directory = os.path.dirname(path)
        if directory:
            mkdir_p(directory)
        return open(path, 'w')

    def _close(self, file): file.close()

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import multiprocessing

from .ProgressReporter import ProgressReporter
from .ProgressReportPickup import ProgressReportPickup

import alphatwirl

##__________________________________________________________________||
class BProgressMonitor(object):
    """A progress monitor of tasks.

    This class monitors the progress of tasks and present it.

    Tasks can be concurrently executed. The progress can be presented
    in an arbitrary way specified at the initialization.

    This class monitors the progress and present it in the background
    by using `ProgressReportPickup`.

    In order for a progress to be monitored by this class, an object
    that executes a task needs to send the `ProgressReport` through
    the `ProgressReporter` created by this class.

    Examples:

      A presentation method needs to be specified at the
      initialization. For example, to use `ProgressBar`::

        presentation = ProgressBar()
        monitor = BProgressMonitor(presentation)

      After the initialization, start monitoring::

        monitor.begin()

      Then, create as many reporters (`ProgressReporter`) as the
      number of the tasks whose progresses need to be monitored::

        reporter1 = monitor.createReporter()
        reporter2 = monitor.createReporter()
        reporter3 = monitor.createReporter()

      These reporters can be given to objects which execute the tasks.
      These objects can be in other processes as long as the reporters
      are properly passed to them.

      If a `ProgressReport` is given to these reporters::

        reporter1.report(report)

      the report will be received by `ProgressReportPickup`, which
      will use the report to present the progress in the way
      specified.

      When all tasks are finished, end monitoring::

        monitor.end()

      This will ends `ProgressReportPickup`, which is running in a
      different process.

    """

    def __init__(self, presentation):
        self.queue = multiprocessing.Queue()
        self.presentation = presentation

    def __repr__(self):
        return '{}(presentation={!r})'.format(
            self.__class__.__name__,
            self.presentation
        )

    def begin(self):
        self.pickup = ProgressReportPickup(self.queue, self.presentation)
        self.pickup.start()
        reporter = self.createReporter()
        alphatwirl.progressbar._progress_reporter = reporter

    def end(self):
        alphatwirl.progressbar._progress_reporter = None
        self.queue.put(None)
        self.pickup.join()

    def createReporter(self):
        return ProgressReporter(queue=self.queue)

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import copy
import logging
from collections import OrderedDict

import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

from alphatwirl.loop import EventDatasetReader, EventLoop

##__________________________________________________________________||
@pytest.fixture()
def eventLoopRunner():
    return mock.Mock(name='eventLoopRunner')

@pytest.fixture()
def reader():
    ret = mock.Mock(name='reader')
    ret.configure_mock(name='reader')
    return ret

@pytest.fixture()
def collector():
    return mock.Mock(name='collector')

@pytest.fixture()
def split_into_build_events():
    return mock.Mock(name='split_into_build_events')

@pytest.fixture()
def obj(eventLoopRunner, reader, collector, split_into_build_events):
    return EventDatasetReader(eventLoopRunner, reader, collector,
                                 split_into_build_events)

##__________________________________________________________________||
def test_dupliate_dataset_name(obj, eventLoopRunner, reader, collector,
                  split_into_build_events, caplog):

    ## begin
    obj.begin()

    ## create data sets
    # dataset1 - 3 event builders
    build_events1 = mock.Mock(name='build_events1')
    build_events2 = mock.Mock(name='build_events2')
    build_events3 = mock.Mock(name='build_events3')
    dataset1 = mock.Mock(name='dataset1', build_events=[build_events1, build_events2, build_events3])
    dataset1.configure_mock(name='dataset1')

    # dataset2 - no event builder
    dataset2 = mock.Mock(name='dataset2', build_events=[ ])
    dataset2.configure_mock(name='dataset2')

    # dataset3 - 1 event builder
    build_events4 = mock.Mock(name='build_events4')
    dataset3 = mock.Mock(name='dataset3', build_events=[build_events4])
    dataset3.configure_mock(name='dataset1') # the same name as dataset1

    eventLoopRunner.run_multiple.side_effect = [[0, 1, 2], [ ], [3]]
    split_into_build_events.side_effect = lambda dataset: dataset.build_events

    ## read
    obj.read(dataset1)
    obj.read(dataset2)
    obj.read(dataset3)

    assert 0 == eventLoopRunner.run.call_count
    assert 3 == eventLoopRunner.run_multiple.call_count

    call1 = eventLoopRunner.run_multiple.call_args_list[0]

    eventLoop1 = call1[0][0][0]
    assert isinstance(eventLoop1, EventLoop)
    assert build_events1 is eventLoop1.build_events
    assert reader is not eventLoop1.reader
    assert 'reader' == eventLoop1.reader.name

    eventLoop2 = call1[0][0][1]
    assert isinstance(eventLoop2, EventLoop)
    assert build_events2 is eventLoop2.build_events
    assert reader is not eventLoop2.reader
    assert 'reader' == eventLoop2.reader.name

    eventLoop3 = call1[0][0][2]
    assert isinstance(eventLoop3, EventLoop)
    assert build_events3 is eventLoop3.build_events
    assert reader is not eventLoop3.reader
    assert 'reader' == eventLoop3.reader.name

    call2 = eventLoopRunner.run_multiple.call_args_list[1]
    assert mock.call([ ]) == call2

    call3 = eventLoopRunner.run_multiple.call_args_list[2]

    eventLoop4 = call3[0][0][0]
    assert isinstance(eventLoop4, EventLoop)
    assert build_events4 is eventLoop4.build_events
    assert reader is not eventLoop4.reader
    assert 'reader' == eventLoop4.reader.name

    ## end
    eventLoopRunner.receive_one.side_effect = [
        (0, eventLoop1.reader), (1, eventLoop2.reader), (2, eventLoop3.reader), (3, eventLoop4.reader)
    ]
    collector.collect.side_effect = lambda x: x
    assert 0 == eventLoopRunner.receive_one.call_count
    assert 0 == collector.collect.call_count
    results = obj.end()
    assert 4 == eventLoopRunner.receive_one.call_count
    assert 1 == collector.collect.call_count
    expected = [
        ('dataset1', (eventLoop1.reader, eventLoop2.reader, eventLoop3.reader), ),
        ('dataset2', ( )),
        ('dataset3', (eventLoop4.reader, )),
    ]

    # results = [
    #     ('dataset1', [deepcopy(reader)]),
    #     ('dataset2', [deepcopy(reader)]),
    #     ('dataset3', [deepcopy(reader)]),
    # ]

    assert 2 == len(results)

    assert 'dataset1' == results[0][0]
    assert eventLoop1.reader == results[0][1][0]
    assert [
        mock.call(eventLoop2.reader),
        mock.call(eventLoop3.reader),
        mock.call(eventLoop4.reader),
    ] == eventLoop1.reader.merge.call_args_list

    assert 'dataset2' == results[1][0]

    assert [
    ] == eventLoop4.reader.merge.call_args_list

##__________________________________________________________________||
